import logging
import pprint

from collections import OrderedDict, defaultdict
from typing import Iterable


from andes.shared import jac_full_names, jac_names, jac_types
from andes.utils.tab import make_doc_table, math_wrap

logger = logging.getLogger(__name__)


class ModelFlags:
    """
    Model flags.

    Parameters
    ----------
    collate : bool
        True: collate variables by device; False: by variable.
        Non-collate (continuous memory) has faster computation speed.
    pflow : bool
        True: called during power flow
    tds : bool
        True if called during tds; if is False, ``dae_t`` cannot be used
    pflow_init : bool or None
        True if initialize pflow; False otherwise; None default to `pflow`
    tds_init : bool or None
        True if initialize tds; False otherwise; None default to `tds`
    series : bool
        True if is series device
    nr_iter : bool
        True if is series device
    f_num : bool
        True if the model defines `f_numeric`
    g_num : bool
        True if the model defines `g_numeric`
    j_num : bool
        True if the model defines `j_numeric`
    s_num : bool
        True if the model defines `s_numeric`
    sv_num : bool
        True if the model defines `s_numeric_var`
    jited : bool
        True if numba JIT code is generated
    """

    def __init__(self, collate=False, pflow=False, tds=False,
                 pflow_init=None, tds_init=None, series=False,
                 nr_iter=False, f_num=False, g_num=False, j_num=False,
                 s_num=False, sv_num=False):

        self.collate = collate
        self.pflow = pflow
        self.tds = tds
        self.pflow_init = pflow_init
        self.tds_init = tds_init
        self.series = series
        self.nr_iter = nr_iter
        self.f_num = f_num
        self.g_num = g_num
        self.j_num = j_num
        self.s_num = s_num
        self.sv_num = sv_num
        self.sys_base = False
        self.address = False
        self.initialized = False
        self.jited = False

    def update(self, dct):
        self.__dict__.update(dct)

    def __repr__(self):
        return pprint.pformat(self.__dict__)


class DummyValue:
    """
    Class for converting a scalar value to a dummy parameter with `name` and `tex_name` fields.

    A DummyValue object can be passed to Block, which utilizes the `name` field to dynamically generate equations.

    Notes
    -----
    Pass a numerical value to the constructor for most use cases, especially when passing as a v-provider.
    """

    def __init__(self, value):
        if isinstance(value, str):
            self.name = f'({value})'
        else:
            self.name = value
        self.tex_name = value
        self.v = value


def dummify(param):
    """
    Dummify scalar parameter and return a DummyValue object. Do nothing for BaseParam instances.

    Parameters
    ----------
    param : float, int, str, BaseParam
        parameter object or scalar value

    Returns
    -------
    DummyValue(param) if param is a scalar; param itself, otherwise.

    """
    if isinstance(param, (int, float, str)):
        return DummyValue(param)
    else:
        return param


class JacTriplet:
    """
    Storage class for Jacobian triplet lists.
    """

    def __init__(self):
        self.ijac = defaultdict(list)
        self.jjac = defaultdict(list)
        self.vjac = defaultdict(list)

    def clear_ijv(self):
        """
        Clear stored triplets for all sparse Jacobian matrices
        """
        for j_full_name in jac_full_names:
            self.ijac[j_full_name] = list()
            self.jjac[j_full_name] = list()
            self.vjac[j_full_name] = list()

    def append_ijv(self, j_full_name, ii, jj, vv):
        """
        Append triplets to the given sparse matrix triplets.

        Parameters
        ----------
        j_full_name : str
            Full name of the sparse Jacobian. If is a constant Jacobian, append 'c' to the Jacobian name.
        ii : array-like
            Row indices
        jj : array-like
            Column indices
        vv : array-like
            Value indices
        """

        self.ijac[j_full_name].append(ii)
        self.jjac[j_full_name].append(jj)
        self.vjac[j_full_name].append(vv)

    def ijv(self, j_full_name):
        """
        Return triplet lists in a tuple in the order or (ii, jj, vv)
        """
        return self.ijac[j_full_name], self.jjac[j_full_name], self.vjac[j_full_name]

    def zip_ijv(self, j_full_name):
        """
        Return a zip iterator in the order of (ii, jj, vv)
        """
        return zip(*self.ijv(j_full_name))

    def merge(self, triplet):
        """
        Merge another triplet into this one.
        """
        for jname in jac_names:
            for jtype in jac_types:
                self.ijac[jname + jtype] += triplet.ijac[jname + jtype]
                self.jjac[jname + jtype] += triplet.jjac[jname + jtype]
                self.vjac[jname + jtype] += triplet.vjac[jname + jtype]


class Config:
    """
    A class for storing system, model and routine configurations.
    """

    def __init__(self, name, dct=None, **kwargs):
        """
        Constructor with a dictionary or keyword arguments
        """
        self._name = name
        self._dict = OrderedDict()
        self._help = OrderedDict()
        self._tex = OrderedDict()
        self._alt = OrderedDict()
        self.add(dct, **kwargs)

    def load(self, config):
        """
        Load from a ConfigParser object, ``config``.
        """
        if config is None:
            return
        if self._name in config:
            config_section = config[self._name]
            self.add(OrderedDict(config_section))

    def add(self, dct=None, **kwargs):
        """
        Add config fields from a dictionary or keyword args.

        Existing configs will NOT be overwritten.
        """
        def warn_upper_case(s):
            if any(x.isupper() for x in s):
                logger.warning("Config fields must be in lower case, found %s", s)

        if dct is not None:
            for s in dct.keys():
                warn_upper_case(s)

            self._add(**dct)

        for s in kwargs.keys():
            warn_upper_case(s)

        self._add(**kwargs)

    def add_extra(self, dest, dct=None, **kwargs):
        """
        Add extra contents for config.

        Parameters
        ----------
        dest : str
            Destination string in `_alt`, `_help` or `_tex`.
        dct : OrderedDict, dict
            key: value pairs
        """

        if dct is not None:
            kwargs.update(dct)
        for key, value in kwargs.items():
            if key not in self.__dict__:
                logger.warning("Config field name %s for %s is invalid.", key, dest)
                continue
            self.__dict__[dest][key] = value

    def _add(self, **kwargs):
        """
        Internal function for adding new config keys.

        This function does not perform input data consistency check.
        """

        for key, val in kwargs.items():
            # skip existing entries that are already loaded (from config files)
            if key in self.__dict__:
                continue

            self._set(key, val)

    def _set(self, key, val):
        """
        Set a pair of key and value to the config dict.

        This function does not perform consistency check on input data and will
        not warn of non-existent fields.
        """

        if isinstance(val, str):
            try:
                val = int(val)
            except ValueError:
                try:
                    val = float(val)
                except ValueError:
                    pass

        self.__dict__[key] = val

    def as_dict(self, refresh=False):
        """
        Return the config fields and values in an ``OrderedDict``.

        Values are cached in `self._dict` unless refreshed.
        """
        if refresh is True or len(self._dict) == 0:
            out = []
            for key, val in self.__dict__.items():
                if not key.startswith('_'):
                    out.append((key, val))
            self._dict = OrderedDict(out)

        return self._dict

    def __repr__(self):
        return pprint.pformat(self.as_dict())

    def doc(self, max_width=78, export='plain', target=False, symbol=True):
        out = ''
        if len(self.as_dict()) == 0:
            return out

        if export == 'rest' and target is True:
            max_width = 0
            model_header = '-' * 80 + '\n'
            out += f'.. _{self._name}:\n\n'
            out += model_header + f'{self._name}\n' + model_header
        else:
            model_header = '\n'
            out += model_header + f'Config Fields in [{self._name}]\n' + model_header

        names, value, info = list(), list(), list()
        alt, tex = list(), list()

        for key in self._dict:
            names.append(key)
            value.append(self._dict[key])
            info.append(self._help[key] if key in self._help else '')
            alt.append(self._alt[key] if key in self._alt else '')
            tex.append(self._tex[key] if key in self._tex else '')

        tex = math_wrap(tex, export=export)

        plain_dict = OrderedDict([('Option', names),
                                  ('Value', value),
                                  ('Info', info),
                                  ('Acceptable values', alt)])
        rest_dict = OrderedDict([('Option', names),
                                 ('Symbol', tex),
                                 ('Value', value),
                                 ('Info', info),
                                 ('Accepted values', alt)])

        if not symbol:
            rest_dict.pop("Symbol")

        out += make_doc_table(title="", max_width=max_width, export=export,
                              plain_dict=plain_dict, rest_dict=rest_dict)
        return out

    def check(self):
        """
        Check the validity of config values.
        """
        for key, val in self.as_dict().items():
            if key not in self._alt:
                continue

            _alt = self._alt[key]
            if not isinstance(_alt, Iterable):
                continue
            if isinstance(_alt, str):
                continue
            if val not in _alt:
                raise ValueError(f"[{self._name}].{key}={val} is not a choice from {_alt}.")

        return True

    @property
    def tex_names(self):
        return self._tex

    def update(self, dct: dict = None, **kwargs):
        """
        Update existing configuration fields.

        Values are checked after the update.
        """

        if dct is not None:
            kwargs.update(dct)

        for key, val in kwargs.items():
            self._set(key, val)

        self.check()



"""
Interoperability with pypowsybl.
"""

import logging
import numpy as np
from functools import wraps

from andes.shared import pd
from andes.shared import pypowsybl as pp

logger = logging.getLogger(__name__)


def require_pypowsybl(f):
    """
    Decorator for functions that require pypowsybl.
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            getattr(pp, '__version__')
        except AttributeError:
            raise ModuleNotFoundError("pypowsybl needs to be manually installed.")

        return f(*args, **kwds)

    return wrapper


@require_pypowsybl
def to_pypowsybl(ss):
    """
    Convert an ANDES system to a pypowsybl network.

    Returns
    -------
    pypowsybl.network.Network

    Parameters
    ----------
    ss : andes.system.System
        The ANDES system to be converted.

    Notes
    -----
    - Only the BUS_BREAKER topology is supported.
    - Each bus has a voltage level named "VL" followed by the bus idx.
    - Buses connected by transformers are in the same substation.

    Warnings
    --------
    - Power flow results are not verified.

    Examples
    --------
    One can utilize pypowsybl to draw network topology. For example,

    .. code:: python

        ss = andes.system.example()
        n = to_pypowsybl(ss)
        results = pp.loadflow.run_ac(n)

        n.get_network_area_diagram()  # show diagram for system
        n.get_single_line_diagram("VL6")  # show single-line diagram for bus 6

    """

    substations, voltage_levels = _make_substation_voltage(ss)
    buses = _make_buses(ss)
    lines, transformers = _make_tline_tf(ss)
    loads = _make_loads(ss)
    generators = _make_generators(ss)
    shunt_df, shunt_model_df = _make_shunts(ss)

    n = pp.network.create_empty()
    n.create_substations(substations)
    n.create_voltage_levels(voltage_levels)
    n.create_buses(buses)
    n.create_lines(lines)
    n.create_loads(loads)
    n.create_generators(generators)
    n.create_shunt_compensators(shunt_df, non_linear_model_df=shunt_model_df)
    n.create_2_windings_transformers(transformers)

    return n


def _make_substation_voltage(ss):
    """
    Make dataframes for substation and voltage levels.

    """
    # --- get `idx` for transmission lines (tline) and transformers (tf)
    tf_idx = ss.Line.get_tf_idx()
    bus1_tf = ss.Line.get("bus1", tf_idx)
    bus2_tf = ss.Line.get("bus2", tf_idx)

    # buses attached to the same transformer will be in one substation

    # -- map bus idx (number or string) -> substation idx (number or string)

    bus2subs = {bus_idx: f"S{bus_idx}" for bus_idx in ss.Bus.idx.v}
    in_substation = {}

    # point buses to substations based on transformers
    for bus1, bus2 in zip(bus1_tf, bus2_tf):
        if (bus1 not in in_substation) and (bus2 not in in_substation):
            bus2subs[bus2] = bus2subs[bus1]
            in_substation[bus1] = in_substation[bus2] = bus2subs[bus1]
        elif (bus1 in in_substation):
            bus2subs[bus2] = bus2subs[bus1] = in_substation[bus1]
            in_substation[bus2] = bus2subs[bus1]
        elif (bus2 in in_substation):
            bus2subs[bus2] = bus2subs[bus1] = in_substation[bus2]
            in_substation[bus1] = bus2subs[bus1]

    # one substation can have multiple voltage levels --- map voltage level
    # (number) -> substation name (string) ---
    vl2subs = {bus_idx: f"S{bus_idx}" for bus_idx in ss.Bus.idx.v}

    # update voltage levels of `bus2` to point to substation named after `bus1`
    for bus1, bus2 in zip(bus1_tf, bus2_tf):
        vl2subs[bus1] = bus2subs[bus1]
        vl2subs[bus2] = bus2subs[bus2]

    substation_ids = sorted(list(set(bus2subs.values())))

    country = 'US'
    substations = pd.DataFrame.from_records(
        index='id',
        data={'id': substation_ids,
              'country': [country] * len(substation_ids),
              },
    )

    # A substation can have multiple voltage levels In out case, one Bus has one
    # voltage level
    voltage_levels = pd.DataFrame.from_records(
        index='id',
        data={"id": ["VL{}".format(item) for item in ss.Bus.idx.v],
              "substation_id": [vl2subs[bus] for bus in ss.Bus.idx.v],
              "topology_kind": "BUS_BREAKER",
              "nominal_v": ss.Bus.Vn.v,
              }
    )

    return substations, voltage_levels


def _make_buses(ss):
    buses = pd.DataFrame.from_records(index='id', data={
        'voltage_level_id': ["VL{}".format(item) for item in ss.Bus.idx.v],
        'id':  ["{}".format(item) for item in ss.Bus.idx.v],
    })

    return buses


def _make_tline_tf(ss):
    """
    Generate line and 2-winding transformer data.

    """

    tf_idx = ss.Line.get_tf_idx()
    tline_idx = ss.Line.get_tline_idx()

    tline_pos = ss.Line.idx2uid(tline_idx)
    tf_pos = ss.Line.idx2uid(tf_idx)

    transformers = pd.DataFrame.from_records(index='id', data={
        'id': ['{}'.format(i) for i in ss.Line.get("idx", tf_idx)],
        'name': ['{}'.format(i) for i in ss.Line.get("name", tf_idx)],
        'voltage_level1_id': ["VL{}".format(item) for item in ss.Line.get("bus1", tf_idx)],
        'voltage_level2_id': ["VL{}".format(item) for item in ss.Line.get("bus2", tf_idx)],
        'bus1_id': ["{}".format(item) for item in ss.Line.get("bus1", tf_idx)],
        'connectable_bus1_id': ["{}".format(item) for item in ss.Line.get("bus1", tf_idx)],
        'bus2_id': ["{}".format(item) for item in ss.Line.get("bus2", tf_idx)],
        'connectable_bus2_id': ["{}".format(item) for item in ss.Line.get("bus2", tf_idx)],
        'rated_u1': ss.Bus.get("Vn", ss.Line.get("bus1", tf_idx)),
        'rated_u2': ss.Bus.get("Vn", ss.Line.get("bus2", tf_idx)),
        'rated_s': 9999 * np.ones(len(tf_idx)),
        'r': ss.Line.get("r", tf_idx) * ss.Line.bases['Zb'][tf_pos],
        'x': ss.Line.get("x", tf_idx) * ss.Line.bases['Zb'][tf_pos],
        'g': ss.Line.get("g", tf_idx) / ss.Line.bases['Zb'][tf_pos],
        'b': ss.Line.get("b", tf_idx) / ss.Line.bases['Zb'][tf_pos],
    })

    lines = pd.DataFrame.from_records(index='id', data={
        'id': ['{}'.format(i) for i in ss.Line.get("idx", tline_idx)],
        'name': ['{}'.format(i) for i in ss.Line.get("name", tline_idx)],
        'voltage_level1_id': ["VL{}".format(item) for item in ss.Line.get("bus1", tline_idx)],
        'voltage_level2_id': ["VL{}".format(item) for item in ss.Line.get("bus2", tline_idx)],
        'bus1_id': ["{}".format(item) for item in ss.Line.get("bus1", tline_idx)],
        'bus2_id': ["{}".format(item) for item in ss.Line.get("bus2", tline_idx)],
        'r': ss.Line.get("r", tline_idx) * ss.Line.bases['Zb'][tline_pos],
        'x': ss.Line.get("x", tline_idx) * ss.Line.bases['Zb'][tline_pos],
        'g1': ss.Line.get("g1", tline_idx) / ss.Line.bases['Zb'][tline_pos],
        'b1': ss.Line.get("b1", tline_idx) / ss.Line.bases['Zb'][tline_pos],
        'g2': ss.Line.get("g2", tline_idx) / ss.Line.bases['Zb'][tline_pos],
        'b2': ss.Line.get("b2", tline_idx) / ss.Line.bases['Zb'][tline_pos],
    })

    return lines, transformers


def _make_loads(ss):

    loads = pd.DataFrame.from_records(index='id', data={
        'voltage_level_id': ["VL{}".format(item) for item in ss.PQ.bus.v],
        'id': ["{}".format(item) for item in ss.PQ.idx.v],
        'bus_id': ["{}".format(item) for item in ss.PQ.bus.v],
        'p0': ss.PQ.p0.v * ss.config.mva,
        'q0': ss.PQ.q0.v * ss.config.mva,
    })

    return loads


def _make_generators(ss):
    generators = pd.DataFrame.from_records(index='id', data={
        'voltage_level_id': [f"VL{item}" for item in ss.PV.bus.v] + [f"VL{item}" for item in ss.Slack.bus.v],
        'id': [f"GEN{item}" for item in ss.PV.idx.v] + [f"GEN{item}" for item in ss.Slack.idx.v],
        'bus_id': [f"{item}" for item in ss.PV.bus.v] + [f"{item}" for item in ss.Slack.bus.v],
        'target_p': np.hstack([ss.PV.p0.v, ss.Slack.p0.v]) * ss.config.mva,
        'min_p': np.hstack([ss.PV.pmin.v, ss.Slack.pmin.v]) * ss.config.mva,
        'max_p': np.hstack([ss.PV.pmax.v, ss.Slack.pmax.v]) * ss.config.mva,
        'target_v': np.hstack([ss.PV.v0.v * ss.PV.Vn.v, ss.Slack.v0.v * ss.Slack.Vn.v]),
        'voltage_regulator_on': True
    })

    return generators


def _make_shunts(ss):
    shunt_df = pd.DataFrame.from_records(index='id', data={
        'id': [f"SHN{item}" for item in ss.Shunt.idx.v],
        'name': [f"{item}" for item in ss.Shunt.name.v],
        'model_type': 'NON_LINEAR',
        'section_count': 1,
        'target_v': ss.Shunt.Vn.v,
        'target_deadband': 0,
        'voltage_level_id': [f'VL{item}' for item in ss.Shunt.bus.v],
        'bus_id': [f'{item}' for item in ss.Shunt.bus.v],
    })

    shunt_model_df = pd.DataFrame.from_records(
        index='id',
        data={"id": [f"SHN{item}" for item in ss.Shunt.idx.v],
              "g": ss.Shunt.g.v / ss.Shunt.bases['Zb'],
              "b": ss.Shunt.b.v / ss.Shunt.bases['Zb'],
              }
    )

    return shunt_df, shunt_model_df



from collections import OrderedDict

from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        IdxParam, Lag, LeadLag, LessThan, Limiter, Model,
                        ModelData, NumParam, Switcher,)
from andes.core.block import DeadBand1, PITrackAW
from andes.core.service import (CurrentSign, DataSelect, DeviceFinder,
                                NumSelect, VarHold, VarService,)


class REPCA1Data(ModelData):
    """
    Parameters for the Renewable Energy Plant Control model.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.ree = IdxParam(info='RenExciter idx',
                            model='RenExciter',
                            mandatory=True,
                            )

        self.line = IdxParam(info='Idx of line that connect to measured bus',
                             model='ACLine',
                             mandatory=True,
                             )

        self.busr = IdxParam(info='Optional remote bus for voltage and freq. measurement',
                             model='Bus',
                             default=None,
                             )

        self.busf = IdxParam(info='BusFreq idx for mode 2',
                             model='BusFreq',
                             default=None,
                             )

        # --- flags ---
        self.VCFlag = NumParam(info='Droop flag; 0-with droop if power factor ctrl, 1-line drop comp.',
                               mandatory=True,
                               unit='bool',
                               )

        self.RefFlag = NumParam(info='Q/V select; 0-Q control, 1-V control',
                                mandatory=True,
                                unit='bool',
                                )

        self.Fflag = NumParam(info='Frequency control flag; 0-disable, 1-enable',
                              mandatory=True,
                              unit='bool',
                              )

        self.PLflag = NumParam(info='Pline ctrl. flag; 0-disable, 1-enable',
                               default=True,
                               unit='bool',
                               )

        self.Tfltr = NumParam(default=0.02,
                              tex_name='T_{fltr}',
                              info='V or Q filter time const.',
                              )

        self.Kp = NumParam(default=1.0,
                           tex_name='K_p',
                           info='Q proportional gain',
                           )

        self.Ki = NumParam(default=0.1,
                           tex_name='K_i',
                           info='Q integral gain',
                           )

        self.Tft = NumParam(default=1.0,
                            tex_name='T_{ft}',
                            info='Lead time constant',
                            )

        self.Tfv = NumParam(default=1.0,
                            tex_name='T_{fv}',
                            info='Lag time constant',
                            )

        self.Vfrz = NumParam(default=0.8,
                             tex_name='V_{frz}',
                             info='Voltage below which s2 is frozen',
                             )

        self.Rc = NumParam(default=None,
                           tex_name='R_c',
                           info='Line drop compensation R',
                           )

        self.Xc = NumParam(default=None,
                           tex_name='X_c',
                           info='Line drop compensation R',
                           )

        self.Kc = NumParam(default=0.0,
                           tex_name='K_c',
                           info='Reactive power compensation gain',
                           )

        self.emax = NumParam(default=999,
                             tex_name='e_{max}',
                             info='Upper limit on deadband output',
                             )

        self.emin = NumParam(default=-999,
                             tex_name='e_{min}',
                             info='Lower limit on deadband output',
                             )

        self.dbd1 = NumParam(default=-0.1,
                             tex_name='d_{bd1}',
                             info='Lower threshold for reactive power control deadband (<=0)',
                             )

        self.dbd2 = NumParam(default=0.1,
                             tex_name='d_{bd2}',
                             info='Upper threshold for reactive power control deadband (>=0)',
                             )

        self.Qmax = NumParam(default=999.0,
                             tex_name='Q_{max}',
                             info='Upper limit on output of V-Q control',
                             )

        self.Qmin = NumParam(default=-999.0,
                             tex_name='Q_{min}',
                             info='Lower limit on output of V-Q control',
                             )

        self.Kpg = NumParam(default=1.0,
                            tex_name='K_{pg}',
                            info='Proportional gain for power control',
                            )

        self.Kig = NumParam(default=0.1,
                            tex_name='K_{ig}',
                            info='Integral gain for power control',
                            )

        self.Tp = NumParam(default=0.02,
                           tex_name='T_p',
                           info='Time constant for P measurement',
                           )

        self.fdbd1 = NumParam(default=-0.0002833,
                              tex_name='f_{dbd1}',
                              info='Lower threshold for freq. error deadband',
                              unit='p.u. (Hz)',
                              )

        self.fdbd2 = NumParam(default=0.0002833,
                              tex_name='f_{dbd2}',
                              info='Upper threshold for freq. error deadband',
                              unit='p.u. (Hz)',
                              )

        self.femax = NumParam(default=0.05,
                              tex_name='f_{emax}',
                              info='Upper limit for freq. error',
                              )

        self.femin = NumParam(default=-0.05,
                              tex_name='f_{emin}',
                              info='Lower limit for freq. error',
                              )

        self.Pmax = NumParam(default=999,
                             tex_name='P_{max}',
                             info='Upper limit on power error (used by PI ctrl.)',
                             unit='p.u. (MW)',
                             power=True,
                             )

        self.Pmin = NumParam(default=-999,
                             tex_name='P_{min}',
                             info='Lower limit on power error (used by PI ctrl.)',
                             unit='p.u. (MW)',
                             power=True,
                             )

        self.Tg = NumParam(default=0.02,
                           tex_name='T_g',
                           info='Power controller lag time constant',
                           )

        self.Ddn = NumParam(default=10,
                            tex_name='D_{dn}',
                            info='Reciprocal of droop for over-freq. conditions',
                            )

        self.Dup = NumParam(default=10,
                            tex_name='D_{up}',
                            info='Reciprocal of droop for under-freq. conditions',
                            )


class REPCA1Model(Model):
    """
    REPCA1 model implementation
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)

        self.group = 'RenPlant'
        self.flags.tds = True

        self.config.add(OrderedDict((('kqs', 2),
                                     ('ksg', 2),
                                     ('freeze', 1),
                                     )))

        self.config.add_extra('_help',
                              kqs='Tracking gain for reactive power PI controller',
                              ksg='Tracking gain for active power PI controller',
                              freeze='Voltage dip freeze flag; 1-enable, 0-disable',
                              )
        self.config.add_extra('_tex',
                              kqs='K_{qs}',
                              ksg='K_{sg}',
                              freeze='f_{rz}')

        # --- from RenExciter ---
        self.reg = ExtParam(model='RenExciter', src='reg', indexer=self.ree, export=False,
                            info='Retrieved RenGen idx', vtype=str, default=None,
                            )
        self.Pext = ExtAlgeb(model='RenExciter', src='Pref', indexer=self.ree,
                             info='Pref from RenExciter renamed as Pext',
                             tex_name='P_{ext}',
                             )

        self.Qext = ExtAlgeb(model='RenExciter', src='Qref', indexer=self.ree,
                             info='Qref from RenExciter renamed as Qext',
                             tex_name='Q_{ext}',
                             )

        # --- from RenGen ---
        self.bus = ExtParam(model='RenGen', src='bus', indexer=self.reg, export=False,
                            info='Retrieved bus idx', vtype=str, default=None,
                            )

        self.buss = DataSelect(self.busr, self.bus, info='selected bus (bus or busr)')

        self.busfreq = DeviceFinder(self.busf, link=self.buss, idx_name='bus',
                                    default_model='BusFreq')

        # from Bus
        self.v = ExtAlgeb(model='Bus', src='v', indexer=self.buss, tex_name='V',
                          info='Bus (or busr, if given) terminal voltage',
                          )

        self.a = ExtAlgeb(model='Bus', src='a', indexer=self.buss, tex_name=r'\theta',
                          info='Bus (or busr, if given) phase angle',
                          )

        self.v0 = ExtService(model='Bus', src='v', indexer=self.buss, tex_name="V_0",
                             info='Initial bus voltage',
                             )

        # from BusFreq
        self.f = ExtAlgeb(model='FreqMeasurement', src='f', indexer=self.busfreq, export=False,
                          info='Bus frequency', unit='p.u.')

        # from Line
        self.bus1 = ExtParam(model='ACLine', src='bus1', indexer=self.line, export=False,
                             info='Retrieved Line.bus1 idx', vtype=str, default=None,
                             )

        self.bus2 = ExtParam(model='ACLine', src='bus2', indexer=self.line, export=False,
                             info='Retrieved Line.bus2 idx', vtype=str, default=None,
                             )
        self.r = ExtParam(model='ACLine', src='r', indexer=self.line, export=False,
                          info='Retrieved Line.r', vtype=str, default=None,
                          )

        self.x = ExtParam(model='ACLine', src='x', indexer=self.line, export=False,
                          info='Retrieved Line.x', vtype=str, default=None,
                          )

        self.v1 = ExtAlgeb(model='ACLine', src='v1', indexer=self.line, tex_name='V_1',
                           info='Voltage at Line.bus1',
                           )

        self.v2 = ExtAlgeb(model='ACLine', src='v2', indexer=self.line, tex_name='V_2',
                           info='Voltage at Line.bus2',
                           )

        self.a1 = ExtAlgeb(model='ACLine', src='a1', indexer=self.line, tex_name=r'\theta_1',
                           info='Angle at Line.bus1',
                           )

        self.a2 = ExtAlgeb(model='ACLine', src='a2', indexer=self.line, tex_name=r'\theta_2',
                           info='Angle at Line.bus2',
                           )

        # -- begin services ---

        self.Isign = CurrentSign(self.bus, self.bus1, self.bus2, tex_name='I_{sign}')

        Iline = '(Isign * (v1*exp(1j*a1) - v2*exp(1j*a2)) / (r + 1j*x))'

        self.Iline = VarService(v_str=Iline, vtype=complex,
                                info='Complex current from bus1 to bus2',
                                tex_name='I_{line}',
                                )

        self.Iline0 = ConstService(v_str='Iline', vtype=complex,
                                   info='Initial complex current from bus1 to bus2',
                                   tex_name='I_{line0}',
                                   )

        Pline = 're(Isign * v1*exp(1j*a1) * conj((v1*exp(1j*a1) - v2*exp(1j*a2)) / (r + 1j*x)))'

        self.Pline = VarService(v_str=Pline, vtype=float,
                                info='Complex power from bus1 to bus2',
                                tex_name='P_{line}',
                                )

        self.Pline0 = ConstService(v_str='Pline', vtype=float,
                                   info='Initial vomplex power from bus1 to bus2',
                                   tex_name='P_{line0}',
                                   )

        Qline = 'im(Isign * v1*exp(1j*a1) * conj((v1*exp(1j*a1) - v2*exp(1j*a2)) / (r + 1j*x)))'

        self.Qline = VarService(v_str=Qline, vtype=float,
                                info='Complex power from bus1 to bus2',
                                tex_name='Q_{line}',
                                )

        self.Qline0 = ConstService(v_str='Qline', vtype=float,
                                   info='Initial complex power from bus1 to bus2',
                                   tex_name='Q_{line0}',
                                   )

        self.Rcs = NumSelect(self.Rc, self.r, info='Line R (Rc if provided, otherwise line.r)',
                             tex_name='R_{cs}',
                             )

        self.Xcs = NumSelect(self.Xc, self.x, info='Line X (Xc if provided, otherwise line.x)',
                             tex_name='X_{cs}',
                             )

        self.Vcomp = VarService(v_str='abs(v*exp(1j*a) - (Rcs + 1j * Xcs) * Iline)',
                                info='Voltage after Rc/Xc compensation',
                                tex_name='V_{comp}'
                                )

        self.SWVC = Switcher(u=self.VCFlag, options=(0, 1), tex_name='SW_{VC}', cache=True)

        self.SWRef = Switcher(u=self.RefFlag, options=(0, 1), tex_name='SW_{Ref}', cache=True)

        self.SWF = Switcher(u=self.Fflag, options=(0, 1), tex_name='SW_{F}', cache=True)

        self.SWPL = Switcher(u=self.PLflag, options=(0, 1), tex_name='SW_{PL}', cache=True)

        VCsel = '(SWVC_s1 * Vcomp + SWVC_s0 * (Qline * Kc + v))'

        self.Vref0 = ConstService(v_str='(SWVC_s1 * Vcomp + SWVC_s0 * (Qline0 * Kc + v))',
                                  tex_name='V_{ref0}',
                                  )

        self.s0 = Lag(VCsel, T=self.Tfltr, K=1, tex_name='s_0',
                      info='V filter',
                      )  # s0_y is the filter output of voltage deviation

        self.s1 = Lag(self.Qline, T=self.Tfltr, K=1, tex_name='s_1')

        self.Vref = Algeb(v_str='Vref0', e_str='Vref0 - Vref', tex_name='Q_{ref}')

        self.Qlinef = Algeb(v_str='Qline0', e_str='Qline0 - Qlinef', tex_name='Q_{linef}')

        Refsel = '(SWRef_s0 * (Qlinef - s1_y) + SWRef_s1 * (Vref - s0_y))'

        self.Refsel = Algeb(v_str=Refsel, e_str=f'{Refsel} - Refsel', tex_name='R_{efsel}')

        self.dbd = DeadBand1(u=self.Refsel, lower=self.dbd1, upper=self.dbd2, center=0.0,
                             tex_name='d^{bd}',
                             )

        # --- e Hardlimit and hold logic ---
        self.eHL = Limiter(u=self.dbd_y, lower=self.emin, upper=self.emax,
                           tex_name='e_{HL}',
                           info='Hardlimit on deadband output',
                           )

        self.zf = VarService(v_str='Indicator(v < Vfrz) * freeze',
                             tex_name='z_f',
                             info='PI Q input freeze signal',
                             )

        self.enf = Algeb(tex_name='e_{nf}',
                         info='e Hardlimit output before freeze',
                         v_str='dbd_y*eHL_zi + emax*eHL_zu + emin*eHL_zl',
                         e_str='dbd_y*eHL_zi + emax*eHL_zu + emin*eHL_zl - enf',
                         )

        # --- hold of `enf` when v < vfrz

        self.eHld = VarHold(u=self.enf, hold=self.zf, tex_name='e_{hld}',
                            info='e Hardlimit output after conditional hold',
                            )

        self.s2 = PITrackAW(u='eHld',
                            kp=self.Kp, ki=self.Ki, ks=self.config.kqs,
                            lower=self.Qmin, upper=self.Qmax,
                            info='PI controller for eHL output',
                            tex_name='s_2',
                            )

        self.s3 = LeadLag(u=self.s2_y, T1=self.Tft, T2=self.Tfv, K=1,
                          tex_name='s_3',
                          )  # s3_y == Qext

        # Active power part

        self.s4 = Lag(self.Pline, T=self.Tp, K=1,
                      tex_name='s_4',
                      info='Pline filter',
                      )

        self.Freq_ref = ConstService(v_str='1.0',
                                     tex_name='f_{ref}',
                                     info='Initial Freq_ref')
        self.ferr = Algeb(tex_name='f_{err}',
                          info='Frequency deviation',
                          unit='p.u. (Hz)',
                          v_str='(Freq_ref - f)',
                          e_str='(Freq_ref - f) - ferr',
                          )

        self.fdbd = DeadBand1(u=self.ferr, center=0.0, lower=self.fdbd1,
                              upper=self.fdbd2,
                              tex_name='f^{dbd}',
                              info='frequency error deadband',
                              )

        self.fdlt0 = LessThan(self.fdbd_y, 0.0,
                              tex_name='f_{dlt0}',
                              info='frequency deadband output less than zero',
                              )

        fdroop = '(fdbd_y * Ddn * fdlt0_z1 + fdbd_y * Dup * fdlt0_z0)'

        self.Plant_pref = Algeb(tex_name='P_{ref}',
                                info='Plant P ref',
                                v_str='Pline0',
                                e_str='Pline0 - Plant_pref',
                                )

        self.Plerr = Algeb(tex_name='P_{lerr}',
                           info='Pline error',
                           v_str='- s4_y + Plant_pref',
                           e_str='- s4_y + Plant_pref - Plerr',
                           )

        self.Perr = Algeb(tex_name='P_{err}',
                          info='Power error before fe limits',
                          v_str=f'{fdroop} + Plerr * SWPL_s1',
                          e_str=f'{fdroop} + Plerr * SWPL_s1 - Perr',
                          )

        self.feHL = Limiter(self.Perr, lower=self.femin, upper=self.femax,
                            tex_name='f_{eHL}',
                            info='Limiter for power (frequency) error',
                            )

        feout = '(Perr * feHL_zi + femin * feHL_zl + femax * feHL_zu)'
        self.s5 = PITrackAW(u=feout, kp=self.Kpg, ki=self.Kig, ks=self.config.ksg,
                            lower=self.Pmin, upper=self.Pmax,
                            tex_name='s_5',
                            info='PI for fe limiter output',
                            )

        self.s6 = Lag(u=self.s5_y, T=self.Tg, K=1,
                      tex_name='s_6',
                      info='Output filter for Pext',
                      )

        Qext = '(s3_y)'

        Pext = '(SWF_s1 * s6_y)'

        self.Pext.e_str = Pext

        self.Qext.e_str = Qext


class REPCA1(REPCA1Data, REPCA1Model):
    """
    REPCA1: renewable energy power plat control model.

    The output of the model, ``Pext`` and ``Qext``,  are the increment signals
    of active and reactive power for the electrical control model.

    Notes for PSS/E DYR parser:

    1. If ICONs M+1 and M+2 are set to 0 when using generator power, an error
       will be thrown by the parser, saying "<REPCA1> cannot retrieve <bus1>
       from <ACLine> using <line>: KeyError('Group <ACLine> does not contain
       device with idx=False')". Manual effort is required to run the converted
       file. In the REPCA1 sheet, provide the idx of a line that connects to the
       RenGen bus.

    2. PSS/E enters ICONs M+3 as a string in single quotes. The pair of single
       quotes need to be removed, or the conversion will fail.
    """

    def __init__(self, system, config):
        REPCA1Data.__init__(self)
        REPCA1Model.__init__(self, system, config)



"""
HYGOV hydro governor model.
"""

from andes.core import Algeb, ConstService, NumParam, State
from andes.core.block import Integrator, Lag, DeadBand1, AntiWindupRate
from andes.core.service import VarService
from andes.models.governor.tgbase import TGBase, TGBaseData


class HYGOVData(TGBaseData):
    """
    HYGOV data
    """

    def __init__(self):
        super().__init__()
        self.R = NumParam(info='Speed regulation gain (mach. base default)',
                          tex_name='R',
                          default=0.05,
                          unit='p.u.',
                          ipower=True,
                          )
        self.r = NumParam(info='Temporary droop (R<r)',
                          tex_name='r',
                          default=1,
                          unit='p.u.',
                          ipower=True,
                          )
        self.GMAX = NumParam(info='Maximum governor response',
                             tex_name='G_{max}',
                             unit='p.u.',
                             default=1.0,
                             power=True,
                             )
        self.GMIN = NumParam(info='Minimum governor response',
                             tex_name='G_{min}',
                             unit='p.u.',
                             default=0.0,
                             power=True,
                             )
        self.VELM = NumParam(info='Gate velocity limit',
                             tex_name='VELM',
                             unit='p.u.',
                             default=0.3,
                             power=True,
                             )

        self.Tf = NumParam(info='Filter time constant',
                           default=0.05,
                           tex_name='T_f')
        self.Tr = NumParam(info='Governor time constant',
                           default=1,
                           tex_name='T_r')
        self.Tg = NumParam(info='Servo time constant',
                           default=0.05,
                           tex_name='T_g')
        self.Dt = NumParam(info='Turbine damping coefficient',
                           default=0.0,
                           tex_name='D_t',
                           power=True,
                           )
        self.qNL = NumParam(info='No-load flow at nominal head',
                            default=0.1,
                            tex_name='q_{NL}',
                            power=True,
                            )
        self.Tw = NumParam(info='Water inertia time constant constant',
                           default=1,
                           tex_name='T_w')
        self.At = NumParam(info='Turbine gain',
                           default=1,
                           tex_name='A_t')

        # nonlinear points, keep for future use
        # self.GV0 = NumParam(info='Governor point 0',
        #                     default=0.1,
        #                     tex_name='GV_0')
        # self.PGV0 = NumParam(info='Governor power point 0',
        #                      default=0,
        #                      tex_name='PGV_0')
        # self.GV1 = NumParam(info='Governor point 1',
        #                     default=0.15,
        #                     tex_name='GV_1')
        # self.PGV1 = NumParam(info='Governor power point 1',
        #                      default=0.1,
        #                      tex_name='PGV_1')
        # self.GV2 = NumParam(info='Governor point 2',
        #                     default=0.3,
        #                     tex_name='GV_2')
        # self.PGV2 = NumParam(info='Governor power point 2',
        #                      default=0.5,
        #                      tex_name='PGV_2')
        # self.GV3 = NumParam(info='Governor point 3',
        #                     default=0.5,
        #                     tex_name='GV_3')
        # self.PGV3 = NumParam(info='Governor power point 3',
        #                      default=0.7,
        #                      tex_name='PGV_3')
        # self.GV4 = NumParam(info='Governor point 4',
        #                     default=0.6,
        #                     tex_name='GV_4')
        # self.PGV4 = NumParam(info='Governor power point 4',
        #                      default=0.8,
        #                      tex_name='PGV_4')
        # self.GV5 = NumParam(info='Governor point 5',
        #                     default=0.7,
        #                     tex_name='GV_5')
        # self.PGV5 = NumParam(info='Governor power point 5',
        #                      default=0.9,
        #                      tex_name='PGV_5')


class HYGOVDBData(HYGOVData):
    """
    HYGOVDB data.
    """

    def __init__(self):
        HYGOVData.__init__(self)
        self.dbL = NumParam(info='Lower bound of deadband',
                            tex_name='db_L',
                            default=0.0,
                            unit='p.u.',
                            )
        self.dbU = NumParam(info='Upper bound of deadband',
                            tex_name='db_U',
                            default=0.0,
                            unit='p.u.',
                            )


class HYGOVModel(TGBase):
    """
    Implement HYGOV model.

    The input lead-lag filter is ignored.

    The ``g`` signal (LAG) is initialized to the initial value of
    ``q`` (Integrator) to simplify the initializaiton equations.
    """

    def __init__(self, system, config):
        TGBase.__init__(self, system, config)

        self.VELMn = ConstService(v_str='-VELM',
                                  tex_name='-VELM',
                                  )
        self.tr = ConstService(v_str='r * Tr',
                               tex_name='r*Tr',
                               )
        self.gr = ConstService(v_str='1/r',
                               tex_name='1/r',
                               )
        self.ratel = ConstService(v_str='- VELM - gr',
                                  tex_name='rate_l',
                                  )
        self.rateu = ConstService(v_str='VELM - gr',
                                  tex_name='rate_u',
                                  )
        self.q0 = ConstService(v_str='tm0 / At + qNL',
                               tex_name='q_0',
                               )
        self.pref = Algeb(info='Reference power input',
                          tex_name='P_{ref}',
                          v_str='R * q0',
                          e_str='R * q0 - pref',
                          )

        self.wd = Algeb(info='Generator speed deviation',
                        unit='p.u.',
                        tex_name=r'\omega_{dev}',
                        v_str='0',
                        e_str='ue * (omega - wref) - wd',
                        )
        self.pd = Algeb(info='Pref plus speed deviation times gain',
                        unit='p.u.',
                        tex_name="P_d",
                        v_str='0',
                        e_str='ue * (- wd + pref + paux - R * dg) - pd',
                        )

        self.LG = Lag(u=self.pd,
                      K=1,
                      T=self.Tf,
                      info='filter after speed deviation (e)',
                      )

        self.gtpos = State(info='State in gate position (c)',
                           unit='rad',
                           v_str='q0',
                           tex_name=r'\delta',
                           e_str='LG_y')

        self.dgl = VarService(tex_name='dg_{lower}', info='dg lower limit',
                              v_str='- VELM - gr * LG_y',)
        self.dgu = VarService(tex_name='dg_{upper}', info='dg upper limit',
                              v_str='VELM - gr * LG_y',)
        self.dg_lim = AntiWindupRate(u=self.gtpos, lower=self.GMIN, upper=self.GMAX,
                                     rate_lower=self.dgl, rate_upper=self.dgu,
                                     tex_name='lim_{dg}',
                                     info='gate velocity and position limiter',
                                     )

        self.dg = Algeb(info='desired gate (c)',
                        unit='p.u.',
                        tex_name="dg",
                        v_str='q0',
                        e_str='gtpos + gr * LG_y - dg',
                        )

        self.LAG = Lag(u=self.dg,
                       K=1,
                       T=self.Tg,
                       info='gate opening (g)',
                       )
        self.h = Algeb(info='turbine head',
                       unit='p.u.',
                       tex_name="h",
                       e_str='q_y**2 / LAG_y**2 - h',
                       v_str='1',
                       )
        self.q = Integrator(u="1 - q_y**2 / LAG_y**2",
                            T=self.Tw, K=1,
                            y0='q0',
                            check_init=False,
                            info="turbine flow (q)"
                            )

        self.pout.e_str = 'ue * (At * h * (q_y - qNL) - Dt * wd * LAG_y) - pout'


class HYGOVDBModel(HYGOVModel):
    """
    Model HYGOV with deadband.
    """

    def __init__(self, system, config):
        HYGOVModel.__init__(self, system, config)
        self.DB = DeadBand1(u=self.wd, center=0.0, lower=self.dbL,
                            upper=self.dbU, tex_name='DB',
                            info='deadband for speed deviation',
                            )
        self.pd.e_str = 'ue * (- DB_y + pref + paux - R * dg) - pd'


class HYGOV(HYGOVData, HYGOVModel):
    """
    HYGOV turbine governor model.

    Implements the PSS/E HYGOV model without deadband.

    Reference:

    [1] PSSE, Model Library, HYGOV
    """

    def __init__(self, system, config):
        HYGOVData.__init__(self)
        HYGOVModel.__init__(self, system, config)


class HYGOVDB(HYGOVDBData, HYGOVDBModel):
    """
    HYGOV turbine governor model with speed input deadband.
    """

    def __init__(self, system, config):
        HYGOVDBData.__init__(self)
        HYGOVDBModel.__init__(self, system, config)



from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        IdxParam, Model, ModelData, NumParam,)
from andes.core.block import Integrator
from andes.core.var import AliasState


class WTDTA1Data(ModelData):
    """
    Data for WTDTA1 wind drive-train model.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.ree = IdxParam(mandatory=True,
                            info='Renewable exciter idx',
                            )

        self.H = NumParam(default=3.0, tex_name='H_t',
                          info='Total inertia constant', unit='MWs/MVA',
                          power=True,
                          non_zero=True,
                          non_negative=True,
                          )

        self.DAMP = NumParam(default=0.0, tex_name='Damp',
                             info='Damp coefficient',
                             unit='p.u. (gen base)',
                             power=True,
                             )

        self.Htfrac = NumParam(default=0.5, tex_name='D_{shaft}',
                               info='Turbine inertia fraction (Hturb/H)',
                               power=True,
                               vrange='[0, 1]',
                               )

        self.Freq1 = NumParam(default=1, tex_name='Freq1',
                              unit='p.u. (Hz)',
                              info='First shaft torsional resonant frequency, p.u. (Hz)',
                              )

        self.Dshaft = NumParam(default=1.0, tex_name='D_{shaft}',
                               info='Shaft damping factor',
                               unit='p.u. (gen base)',
                               power=True,
                               )

        self.w0 = NumParam(default=1.0, tex_name=r'\omega_0',
                           info='Default speed if not using a torque model',
                           unit='p.u.',
                           )


class WTDTA1Model(Model):
    """
    WTDTA1 model equations
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)

        self.flags.tds = True
        self.group = 'RenGovernor'

        self.reg = ExtParam(model='RenExciter', src='reg', indexer=self.ree, vtype=str,
                            export=False,
                            )
        self.Sn = ExtParam(model='RenGen', src='Sn', indexer=self.reg,
                           tex_name='S_n', export=False,
                           )

        self.wge = ExtAlgeb(model='RenExciter', src='wg', indexer=self.ree,
                            export=False,
                            e_str='-1.0 + s2_y',
                            ename='wg',
                            tex_ename=r'\omega_g',
                            )

        self.Pe = ExtAlgeb(model='RenGen', src='Pe', indexer=self.reg, export=False,
                           info='Retrieved Pe of RenGen')

        self.Pe0 = ExtService(model='RenGen', src='Pe', indexer=self.reg, tex_name='P_{e0}',
                              )

        self.Ht2 = ConstService(v_str='2 * (Htfrac * H)', tex_name='2H_t')

        self.Hg2 = ConstService(v_str='2 * H * (1 - Htfrac)', tex_name='2H_g')

        # (2*pi*Freq1)**2 is considered in p.u., which is Freq1**2 here
        self.Kshaft = ConstService(v_str='Ht2 * Hg2 * 0.5 * Freq1 * Freq1 / H', tex_name='K_{shaft}')

        self.wr0 = Algeb(tex_name=r'\omega_{r0}',
                         unit='p.u.',
                         v_str='w0',
                         e_str='w0 - wr0',
                         info='speed set point',
                         )

        self.Pm = Algeb(tex_name='P_m',
                        info='Mechanical power',
                        e_str='Pe0 - Pm',
                        v_str='Pe0',
                        )

        # `s1_y` is `wt`
        self.s1 = Integrator(u='(Pm / s1_y) - s3_y - pd',
                             T=self.Ht2,
                             K=1.0,
                             y0='wr0',
                             )

        self.wt = AliasState(self.s1_y, tex_name=r'\omega_t')

        # `s2_y` is `wg`
        self.s2 = Integrator(u='-(Pe / s2_y) + s3_y - DAMP * (s2_y - w0) + pd',
                             T=self.Hg2,
                             K=1.0,
                             y0='wr0',
                             )

        self.wg = AliasState(self.s2_y, tex_name=r'\omega_g')

        # `s3_y` gets reinitialized in `WTTQA1`
        self.s3 = Integrator(u='s1_y - s2_y',
                             T=1.0,
                             K=self.Kshaft,
                             y0='Pe0 / wr0',
                             )

        self.pd = Algeb(tex_name='P_d', info='Output after damping',
                        e_str='Dshaft * (s1_y - s2_y) - pd',
                        v_str='0',
                        )


class WTDTA1(WTDTA1Data, WTDTA1Model):
    """
    WTDTA wind turbine drive-train model.

    One can set ``Htfrac`` to ``0`` to simulate a single-mass
    drive train. ``Htfrac`` has to be within ``[0, 1]``

    User-provided reference speed should be specified in parameter `w0`.
    Internally, `w0` is set to the algebraic variable `wr0`.

    Note for PSS/E dyr parser:

    In PSS/E doc, `Freq1` is said to be Hz,
    but exported data from PSS/E 34 uses per unit.
    ANDES requires ``Freq1`` in per unit frequency.
    """

    def __init__(self, system, config):
        WTDTA1Data.__init__(self)
        WTDTA1Model.__init__(self, system, config)



from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        ExtState, IdxParam, Model, ModelData, NumParam,)
from andes.core.service import NumSelect


class TGBaseData(ModelData):
    """
    Base data for turbine governors.
    """

    def __init__(self):
        super().__init__()
        self.syn = IdxParam(model='SynGen',
                            info='Synchronous generator idx',
                            mandatory=True,
                            unique=True,
                            )
        self.Tn = NumParam(info='Turbine power rating. Equal to `Sn` if not provided.',
                           tex_name='T_n',
                           unit='MVA',
                           default=None,
                           )
        self.wref0 = NumParam(info='Base speed reference',
                              tex_name=r'\omega_{ref0}',
                              default=1.0,
                              unit='p.u.',
                              )


class TGBase(Model):
    """
    Base Turbine Governor model.

    Parameters
    ----------
    add_sn : bool
        True to add ``NumSelect`` Sn; False to add later in custom models.
        This is useful when the governor connects to two generators.
    add_tm0 : bool
        True to add ``ExtService`` ``tm0``.

    """

    def __init__(self, system, config, add_sn=True, add_tm0=True):
        Model.__init__(self, system, config)
        self.group = 'TurbineGov'
        self.flags.update({'tds': True})
        self.Sg = ExtParam(src='Sn',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name='S_n',
                           info='Rated power from generator',
                           unit='MVA',
                           export=False,
                           )
        self.ug = ExtParam(src='u',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name='u_g',
                           info='Generator connection status',
                           unit='bool',
                           export=False,
                           )
        self.ue = ConstService(v_str='u * ug',
                               info="effective connection status considering generator's",
                               tex_name='u_{e}',
                               )

        if add_sn is True:
            self.Sn = NumSelect(self.Tn,
                                fallback=self.Sg,
                                tex_name='S_n',
                                info='Turbine or Gen rating',
                                )

        self.Vn = ExtParam(src='Vn',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name='V_n',
                           info='Rated voltage from generator',
                           unit='kV',
                           export=False,
                           )

        # Note: changing the values of `tm0` is not allowed at any time!!
        if add_tm0 is True:
            self.tm0 = ExtService(src='tm',
                                  model='SynGen',
                                  indexer=self.syn,
                                  tex_name=r'\tau_{m0}',
                                  info='Initial mechanical input')
            self.pref0 = ConstService(v_str='tm0', info='initial pref',
                                      tex_name='P_{ref0}',
                                      )

        self.omega = ExtState(src='omega',
                              model='SynGen',
                              indexer=self.syn,
                              tex_name=r'\omega',
                              info='Generator speed',
                              unit='p.u.'
                              )

        # Note: changing `paux0` is allowed.
        # It is a way how one can input from external programs such as reinforcement learning.
        self.paux0 = ConstService(v_str='0',
                                  tex_name='P_{aux0}',
                                  info='const. auxiliary input')

        self.tm = ExtAlgeb(src='tm',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'\tau_m',
                           e_str='ue * (pout - tm0)',
                           info='Mechanical power interface to SynGen',
                           ename='tm',
                           tex_ename=r'\tau_{m}',
                           )
        # `paux` must be zero upon initialization
        self.paux = Algeb(info='Auxiliary power input',
                          tex_name='P_{aux}',
                          v_str='paux0',
                          e_str='paux0 - paux',
                          )
        self.pout = Algeb(info='Turbine final output power',
                          tex_name='P_{out}',
                          v_str='ue * tm0',
                          )
        self.wref = Algeb(info='Speed reference variable',
                          tex_name=r'\omega_{ref}',
                          v_str='wref0',
                          e_str='wref0 - wref',
                          )



"""
Three-phase symmetric shunt model.
"""

from andes.core import Algeb
from andes.models.shunt import Shunt


class ShuntTD(Shunt):
    """
    Static shunt model with inverse transformation from phasor to time-domain.
    """

    def __init__(self, system=None, config=None):
        Shunt.__init__(self, system, config)
        self.vta = Algeb(e_str='v / sqrt(3) * cos(2*pi * sys_f * dae_t + a) - vta',
                         v_str='v / sqrt(3) * cos(2*pi * sys_f * dae_t + a)',
                         tex_name='V_{ta}',
                         )
        self.vtb = Algeb(e_str='v / sqrt(3) * cos(2*pi * sys_f * dae_t + a - 2/3*pi) - vtb',
                         v_str='v / sqrt(3) * cos(2*pi * sys_f * dae_t + a - 2/3*pi)',
                         tex_name='V_{tb}',
                         )
        self.vtc = Algeb(e_str='v / sqrt(3) * cos(2*pi * sys_f * dae_t + a + 2/3*pi) - vtc',
                         v_str='v / sqrt(3) * cos(2*pi * sys_f * dae_t + a + 2/3*pi)',
                         tex_name='V_{tc}',
                         )



"""
Round-rotor generator model.
"""

import logging

from andes.core.service import VarService
from andes.models.synchronous.genbase import GENBase, Flux0
from andes.models.synchronous.genrou import GENROUData, GENROUModel

logger = logging.getLogger(__name__)


class GENROUOSModel(GENROUModel):
    def __init__(self):
        GENROUModel.__init__(self)
        delattr(self, 'psi2q')
        delattr(self, 'psi2d')
        delattr(self, 'psi2')

        self.algebs.pop('psi2q')
        self.algebs.pop('psi2d')
        self.algebs.pop('psi2')

        self.psi2q = VarService(tex_name=r"\psi_{aq}", info='q-axis air gap flux',
                                v_str='gq1*e1d + (1-gq1)*e2q',
                                )

        self.psi2d = VarService(tex_name=r"\psi_{ad}", info='d-axis air gap flux',
                                v_str='gd1*e1q + gd2*(xd1-xl)*e2d',
                                )

        self.psi2 = VarService(tex_name=r"\psi_a", info='air gap flux magnitude',
                               v_str='sqrt(psi2d **2 + psi2q ** 2)',
                               )
        # fix the reference to `psi2`
        self.SL.u = self.psi2


class GENROUOS(GENROUData, GENBase, GENROUOSModel, Flux0):
    """
    Round rotor generator with quadratic saturation. The model implements
    operator splitting for the flux linkage equations.
    """

    def __init__(self, system, config):
        GENROUData.__init__(self)
        GENBase.__init__(self, system, config)
        Flux0.__init__(self)
        GENROUOSModel.__init__(self)



"""
Module for renewable energy generator (converter) model A.
"""

from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        IdxParam, Lag, Model, ModelData, NumParam, Piecewise,)
from andes.core.block import GainLimiter, LagAntiWindupRate


class REGCA1Data(ModelData):
    """
    REGC_A model data.

    12/09/2020: Made `Lvplsw` a parameter.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.bus = IdxParam(model='Bus',
                            info="interface bus id",
                            mandatory=True,
                            )
        self.gen = IdxParam(info="static generator index",
                            mandatory=True,
                            )
        self.Sn = NumParam(default=100.0, tex_name='S_n',
                           info='Model MVA base',
                           unit='MVA',
                           )

        self.Tg = NumParam(default=0.1, tex_name='T_g',
                           info='converter time const.', unit='s',
                           )
        self.Rrpwr = NumParam(default=10.0, tex_name='R_{rpwr}',
                              info='Low voltage power logic (LVPL) ramp limit',
                              unit='p.u.',
                              )
        self.Brkpt = NumParam(default=1.0, tex_name='B_{rkpt}',
                              info='LVPL characteristic voltage 2',
                              unit='p.u.',
                              )
        self.Zerox = NumParam(default=0.5, tex_name='Z_{erox}',
                              info='LVPL characteristic voltage 1',
                              unit='p.u',
                              )
        # TODO: ensure Brkpt > Zerox
        self.Lvplsw = NumParam(default=1.0, tex_name='z_{Lvplsw}',
                               info='Low volt. P logic: 1-enable, 0-disable',
                               unit='bool',
                               )

        self.Lvpl1 = NumParam(default=1.0, tex_name='L_{vpl1}',
                              info='LVPL gain',
                              unit='p.u',
                              )
        self.Volim = NumParam(default=1.2, tex_name='V_{olim}',
                              info='Voltage lim for high volt. reactive current mgnt.',
                              unit='p.u.',
                              )
        self.Lvpnt1 = NumParam(default=0.8, tex_name='L_{vpnt1}',
                               info='High volt. point for low volt. active current mgnt.',
                               unit='p.u.',
                               )
        self.Lvpnt0 = NumParam(default=0.4, tex_name='L_{vpnt0}',
                               info='Low volt. point for low volt. active current mgnt.',
                               unit='p.u.',
                               )
        # TODO: ensure Lvpnt1 > Lvpnt0
        self.Iolim = NumParam(default=-1.5, tex_name='I_{olim}',
                              info='lower current limit for high volt. reactive current mgnt.',
                              unit='p.u. (mach base)',
                              current=True,
                              )
        self.Tfltr = NumParam(default=0.1, tex_name='T_{fltr}',
                              info='Voltage filter T const for low volt. active current mgnt.',
                              unit='s',
                              )
        self.Khv = NumParam(default=0.7, tex_name='K_{hv}',
                            info='Overvolt. compensation gain in high volt. reactive current mgnt.',
                            )
        self.Iqrmax = NumParam(default=1, tex_name='I_{qrmax}',
                               info='Upper limit on the ROC for reactive current',
                               unit='p.u.',
                               current=True,
                               )
        self.Iqrmin = NumParam(default=-1, tex_name='I_{qrmin}',
                               info='Lower limit on the ROC for reactive current',
                               unit='p.u.',
                               current=True,
                               )
        self.Accel = NumParam(default=0.0, tex_name='A_{ccel}',
                              info='Acceleration factor',
                              vrange=(0, 1.0),
                              )
        self.gammap = NumParam(default=1.0,
                               info="P ratio of linked static gen",
                               tex_name=r'\gamma_P'
                               )
        self.gammaq = NumParam(default=1.0,
                               info="Q ratio of linked static gen",
                               tex_name=r'\gamma_Q'
                               )


class REGCA1Model(Model):
    """
    REGCA1 implementation.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.flags.tds = True
        self.group = 'RenGen'

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle',
                          e_str='-Pe',
                          ename='P',
                          tex_ename='P',
                          )

        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          info='Bus voltage magnitude',
                          e_str='-Qe',
                          ename='Q',
                          tex_ename='Q',
                          )

        self.p0s = ExtService(model='StaticGen',
                              src='p',
                              indexer=self.gen,
                              tex_name='P_{0s}',
                              info='initial P of the static gen',
                              )
        self.q0s = ExtService(model='StaticGen',
                              src='q',
                              indexer=self.gen,
                              tex_name='Q_{0s}',
                              info='initial Q of the static gen',
                              )
        self.p0 = ConstService(v_str='p0s * gammap',
                               tex_name='P_0',
                               info='initial P of this gen',
                               )
        self.q0 = ConstService(v_str='q0s * gammaq',
                               tex_name='Q_0',
                               info='initial Q of this gen',
                               )
        self.ra = ExtParam(model='StaticGen',
                           src='ra',
                           indexer=self.gen,
                           tex_name='r_a',
                           export=False,
                           )
        self.xs = ExtParam(model='StaticGen',
                           src='xs',
                           indexer=self.gen,
                           tex_name='x_s',
                           export=False,
                           )

        # --- INITIALIZATION ---
        self.q0gt0 = ConstService('Indicator(q0> 0)', tex_name='z_{q0>0}',
                                  info='flags for q0 below zero',
                                  )
        self.q0lt0 = ConstService('Indicator(q0< 0)', tex_name='z_{q0<0}',
                                  info='flags for q0 below zero',
                                  )

        self.Ipcmd0 = ConstService('p0 / v', info='initial Ipcmd',
                                   tex_name='I_{pcmd0}',
                                   )

        self.Iqcmd0 = ConstService('-q0 / v', info='initial Iqcmd',
                                   tex_name='I_{qcmd0}',
                                   )

        self.LVG = Piecewise(u=self.v, points=('Lvpnt0', 'Lvpnt1'),
                             funs=('0', '(v - Lvpnt0) * kLVG', '1'),
                             info='Ip gain during low voltage',
                             tex_name='L_{VG}',
                             )

        # `Ipcmd` is not defined when the initial `LVG_y` is zero
        self.Ipcmd = Algeb(tex_name='I_{pcmd}',
                           info='current component for active power',
                           e_str='Ipcmd0 - Ipcmd * LVG_y',
                           v_str='Indicator(LVG_y>0) * Ipcmd0 / LVG_y + Indicator(LVG_y<=0) * 1',
                           diag_eps=True,
                           )

        self.Iqcmd = Algeb(tex_name='I_{qcmd}',
                           info='current component for reactive power',
                           e_str='Iqcmd0 - Iqcmd', v_str='Iqcmd0')

        # reactive power management

        # rate limiting logic (for fault recovery, although it does not detect any recovery)
        #   - activate upper limit when q0 > 0 (self.q0gt0)
        #   - activate lower limit when q0 < 0 (self.q0lt0)

        self.S1 = LagAntiWindupRate(u=self.Iqcmd,
                                    T=self.Tg, K=-1,
                                    lower=-9999, upper=9999, no_lower=True, no_upper=True,
                                    rate_lower=self.Iqrmin, rate_upper=self.Iqrmax,
                                    rate_lower_cond=self.q0lt0, rate_upper_cond=self.q0gt0,
                                    tex_name='S_1',
                                    info='Iqcmd delay',
                                    )  # output `S1_y` == `Iq`

        # piece-wise gain for low voltage active current mgnt.
        self.kLVG = ConstService(v_str='1 / (Lvpnt1 - Lvpnt0)',
                                 tex_name='k_{LVG}',
                                 )

        # piece-wise gain for LVPL
        self.kLVPL = ConstService(v_str='Lvplsw * Lvpl1 / (Brkpt - Zerox)',
                                  tex_name='k_{LVPL}',
                                  )

        self.S2 = Lag(u=self.v, T=self.Tfltr, K=1.0,
                      info='Voltage filter with no anti-windup',
                      tex_name='S_2',
                      )
        self.LVPL = Piecewise(u=self.S2_y,
                              points=('Zerox', 'Brkpt'),
                              funs=('0 + 9999*(1-Lvplsw)',
                                    '(S2_y - Zerox) * kLVPL + 9999 * (1-Lvplsw)',
                                    '9999'),
                              info='Low voltage Ipcmd upper limit',
                              tex_name='L_{VPL}',
                              )

        self.S0 = LagAntiWindupRate(u=self.Ipcmd, T=self.Tg, K=1,
                                    upper=self.LVPL_y, rate_upper=self.Rrpwr,
                                    lower=-9999, rate_lower=-9999,
                                    no_lower=True, rate_no_lower=True,
                                    tex_name='S_0',
                                    )  # `S0_y` is the output `Ip` in the block diagram

        self.Ipout = Algeb(e_str='S0_y * LVG_y -Ipout',
                           v_str='Ipcmd * LVG_y',
                           info='Output Ip current',
                           tex_name='I_{pout}',
                           )

        # high voltage part
        self.HVG = GainLimiter(u='v - Volim',
                               K=self.Khv, R=1,
                               info='High voltage gain block',
                               lower=0, upper=999, no_upper=True,
                               tex_name='H_{VG}'
                               )
        self.HVG.lim.no_warn = True
        self.HVG.lim.allow_adjust = False

        self.Iqout = GainLimiter(u='S1_y - HVG_y',
                                 K=1, R=1,
                                 lower=self.Iolim, upper=9999,
                                 no_upper=True, info='Iq output block',
                                 tex_name='I^{qout}',
                                 )  # `Iqout_y` is the final Iq output

        self.Pe = Algeb(tex_name='P_e', info='Active power output',
                        v_str='p0', e_str='Ipout * v - Pe')
        self.Qe = Algeb(tex_name='Q_e', info='Reactive power output',
                        v_str='q0', e_str='Iqout_y * v - Qe')

    def v_numeric(self, **kwargs):
        """
        Disable the corresponding `StaticGen`s.
        """
        self.system.groups['StaticGen'].set(src='u', idx=self.gen.v, attr='v', value=0)


class REGCA1(REGCA1Data, REGCA1Model):
    """
    Renewable energy generator model type A.

    Implements ``REGCA1`` in PSS/E, or ``REGC_A`` in PSLF.

    Volim is the voltage limit for high voltage reactive current management,
    which should be large than static bus voltage (Volim > v),
    or initialization error will occur.
    """

    def __init__(self, system, config):
        REGCA1Data.__init__(self)
        REGCA1Model.__init__(self, system, config)



from collections import OrderedDict

import numpy as np

from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        IdxParam, Lag, Limiter, Model, ModelData, NumParam,
                        Piecewise, Switcher, IsEqual)
from andes.core.block import (DeadBand1, GainLimiter, LagAWFreeze, LagFreeze,
                              LagRate, PITrackAWFreeze,)
from andes.core.service import (ApplyFunc, DataSelect, ExtendedEvent, Replace,
                                VarHold, VarService,)
from andes.core.var import AliasState


class REECA1Data(ModelData):
    """
    Renewable energy electrical control model REECA1 (reec_a) data.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.reg = IdxParam(model='RenGen',
                            info='Renewable generator idx',
                            mandatory=True,
                            )

        self.busr = IdxParam(info='Optional remote bus for voltage control',
                             model='Bus',
                             default=None,
                             )
        self.PFFLAG = NumParam(info='Power factor control flag; 1-PF control, 0-Q control',
                               mandatory=True,
                               unit='bool',
                               )
        self.VFLAG = NumParam(info='Voltage control flag; 1-Q control, 0-V control',
                              mandatory=True,
                              unit='bool',
                              )
        self.QFLAG = NumParam(info='Q control flag; 1-V or Q control, 0-const. PF or Q',
                              mandatory=True,
                              unit='bool',
                              )
        self.PFLAG = NumParam(info='P speed-dependency flag; 1-has speed dep., 0-no dep.',
                              mandatory=True,
                              unit='bool',
                              )
        self.PQFLAG = NumParam(info='P/Q priority flag for I limit; 0-Q priority, 1-P priority',
                               mandatory=True,
                               unit='bool',
                               )

        self.Vdip = NumParam(default=0.8,
                             tex_name='V_{dip}',
                             info='Low V threshold to activate Iqinj logic',
                             unit='p.u.',
                             )
        self.Vup = NumParam(default=1.2,
                            tex_name='V_{up}',
                            info='V threshold above which to activate Iqinj logic',
                            unit='p.u.',
                            )
        self.Trv = NumParam(default=0.02,
                            tex_name='T_{rv}',
                            info='Voltage filter time constant',
                            )
        self.dbd1 = NumParam(default=-0.02,
                             tex_name='d_{bd1}',
                             info='Lower bound of the voltage deadband (<=0)',
                             )
        self.dbd2 = NumParam(default=0.02,
                             tex_name='d_{bd2}',
                             info='Upper bound of the voltage deadband (>=0)',
                             )
        self.Kqv = NumParam(default=1.0,
                            vrange=(0, 10),
                            tex_name='K_{qv}',
                            info='Gain to compute Iqinj from V error (caution!!)',
                            )
        self.Iqh1 = NumParam(default=999.0,
                             tex_name='I_{qh1}',
                             info='Upper limit on Iqinj',
                             )
        self.Iql1 = NumParam(default=-999.0,
                             tex_name='I_{ql1}',
                             info='Lower limit on Iqinj',
                             )
        self.Vref0 = NumParam(default=1.0,
                              tex_name='V_{ref0}',
                              info='User defined Vref (if 0, use initial bus V)',
                              )
        self.Iqfrz = NumParam(default=0.0,
                              tex_name='I_{qfrz}',
                              info='Hold Iqinj at the value for Thld (>0) seconds following a Vdip',
                              )
        self.Thld = NumParam(default=0.0,
                             tex_name='T_{hld}',
                             unit='s',
                             info='Time for which Iqinj is held. Hold at Iqinj if>0; hold at State 1 if<0',
                             )
        self.Thld2 = NumParam(default=0.0,
                              tex_name='T_{hld2}',
                              unit='s',
                              info='Time for which IPMAX is held after voltage dip ends',
                              )
        self.Tp = NumParam(default=0.02,
                           tex_name='T_p',
                           unit='s',
                           info='Filter time constant for Pe',
                           )
        self.QMax = NumParam(default=999.0,
                             tex_name='Q_{max}',
                             info='Upper limit for reactive power regulator',
                             )
        self.QMin = NumParam(default=-999.0,
                             tex_name='Q_{min}',
                             info='Lower limit for reactive power regulator',
                             )
        self.VMAX = NumParam(default=999.0,
                             tex_name='V_{max}',
                             info='Upper limit for voltage control',
                             )
        self.VMIN = NumParam(default=-999.0,
                             tex_name='V_{min}',
                             info='Lower limit for voltage control',
                             )
        self.Kqp = NumParam(default=1.0,
                            tex_name='K_{qp}',
                            info='Proportional gain for reactive power error',
                            )
        self.Kqi = NumParam(default=0.1,
                            tex_name='K_{qi}',
                            info='Integral gain for reactive power error',
                            )
        self.Kvp = NumParam(default=1.0,
                            tex_name='K_{vp}',
                            info='Proportional gain for voltage error',
                            )
        self.Kvi = NumParam(default=0.1,
                            tex_name='K_{vi}',
                            info='Integral gain for voltage error',
                            )
        self.Vref1 = NumParam(default=1.0,
                              non_zero=True,
                              tex_name='V_{ref1}',
                              info='Voltage ref. if VFLAG=0',
                              )
        self.Tiq = NumParam(default=0.02,
                            tex_name='T_{iq}',
                            info='Filter time constant for Iq',
                            )
        self.dPmax = NumParam(default=999.0,
                              tex_name='d_{Pmax}',
                              info='Power reference max. ramp rate (>0)',
                              )
        self.dPmin = NumParam(default=-999.0,
                              tex_name='d_{Pin}',
                              info='Power reference min. ramp rate (<0)',
                              )
        self.PMAX = NumParam(default=999.0,
                             tex_name='P_{max}',
                             info='Max. active power limit > 0',
                             )
        self.PMIN = NumParam(default=0.0,
                             tex_name='P_{min}',
                             info='Min. active power limit',
                             )
        self.Imax = NumParam(default=999.0,
                             tex_name='I_{max}',
                             info='Max. apparent current limit',
                             current=True,
                             )
        self.Tpord = NumParam(default=0.02,
                              tex_name='T_{pord}',
                              info='Filter time constant for power setpoint',
                              )
        self.Vq1 = NumParam(default=0.2,
                            tex_name='V_{q1}',
                            info='Reactive power V-I pair (point 1), voltage',
                            )
        self.Iq1 = NumParam(default=2.0,
                            tex_name='I_{q1}',
                            info='Reactive power V-I pair (point 1), current',
                            current=True,
                            )
        self.Vq2 = NumParam(default=0.4,
                            tex_name='V_{q2}',
                            info='Reactive power V-I pair (point 2), voltage',
                            )
        self.Iq2 = NumParam(default=4.0,
                            tex_name='I_{q2}',
                            info='Reactive power V-I pair (point 2), current',
                            current=True,
                            )
        self.Vq3 = NumParam(default=0.8,
                            tex_name='V_{q3}',
                            info='Reactive power V-I pair (point 3), voltage',
                            )
        self.Iq3 = NumParam(default=8.0,
                            tex_name='I_{q3}',
                            info='Reactive power V-I pair (point 3), current',
                            current=True,
                            )
        self.Vq4 = NumParam(default=1.0,
                            tex_name='V_{q4}',
                            info='Reactive power V-I pair (point 4), voltage',
                            )
        self.Iq4 = NumParam(default=10,
                            tex_name='I_{q4}',
                            info='Reactive power V-I pair (point 4), current',
                            current=True,
                            )
        self.Vp1 = NumParam(default=0.2,
                            tex_name='V_{p1}',
                            info='Active power V-I pair (point 1), voltage',
                            )
        self.Ip1 = NumParam(default=2.0,
                            tex_name='I_{p1}',
                            info='Active power V-I pair (point 1), current',
                            current=True,
                            )
        self.Vp2 = NumParam(default=0.4,
                            tex_name='V_{p2}',
                            info='Active power V-I pair (point 2), voltage',
                            )
        self.Ip2 = NumParam(default=4.0,
                            tex_name='I_{p2}',
                            info='Active power V-I pair (point 2), current',
                            current=True,
                            )
        self.Vp3 = NumParam(default=0.8,
                            tex_name='V_{p3}',
                            info='Active power V-I pair (point 3), voltage',
                            )
        self.Ip3 = NumParam(default=8.0,
                            tex_name='I_{p3}',
                            info='Active power V-I pair (point 3), current',
                            current=True,
                            )
        self.Vp4 = NumParam(default=1.0,
                            tex_name='V_{p4}',
                            info='Active power V-I pair (point 4), voltage',
                            )
        self.Ip4 = NumParam(default=12.0,
                            tex_name='I_{p4}',
                            info='Active power V-I pair (point 4), current',
                            current=True,
                            )


class REECA1Model(Model):
    """
    REEC_A model implementation.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)

        self.flags.tds = True
        self.group = 'RenExciter'

        self.config.add(OrderedDict((('kqs', 2),
                                     ('kvs', 2),
                                     ('tpfilt', 0.02),
                                     )))
        self.config.add_extra('_help',
                              kqs='Q PI controller tracking gain',
                              kvs='Voltage PI controller tracking gain',
                              tpfilt='Time const. for Pref filter',
                              )
        self.config.add_extra('_tex',
                              kqs='K_{qs}',
                              kvs='K_{vs}',
                              tpfilt='T_{pfilt}',
                              )

        # --- Sanitize inputs ---
        self.Imaxr = Replace(self.Imax, flt=lambda x: np.less_equal(x, 0), new_val=1e8,
                             tex_name='I_{maxr}')

        # --- Flag switchers ---
        self.SWPF = Switcher(u=self.PFFLAG, options=(0, 1), tex_name='SW_{PF}', cache=True)

        self.SWV = Switcher(u=self.VFLAG, options=(0, 1), tex_name='SW_{V}', cache=True)

        self.SWQ = Switcher(u=self.QFLAG, options=(0, 1), tex_name='SW_{V}', cache=True)

        self.SWP = Switcher(u=self.PFLAG, options=(0, 1), tex_name='SW_{P}', cache=True)

        self.SWPQ = Switcher(u=self.PQFLAG, options=(0, 1), tex_name='SW_{PQ}', cache=True)

        # --- External parameters ---
        self.bus = ExtParam(model='RenGen', src='bus', indexer=self.reg, export=False,
                            info='Retrieved bus idx', vtype=str, default=None,
                            )

        self.buss = DataSelect(self.busr, self.bus, info='selected bus (bus or busr)')

        self.gen = ExtParam(model='RenGen', src='gen', indexer=self.reg, export=False,
                            info='Retrieved StaticGen idx', vtype=str, default=None,
                            )

        self.Sn = ExtParam(model='RenGen', src='Sn', indexer=self.reg,
                           tex_name='S_n', export=False,
                           )

        # --- External variables ---
        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle',
                          )

        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          info='Bus voltage magnitude',
                          )  # check whether to use `bus` or `buss`

        self.Pe = ExtAlgeb(model='RenGen', src='Pe', indexer=self.reg, export=False,
                           info='Retrieved Pe of RenGen')

        self.Qe = ExtAlgeb(model='RenGen', src='Qe', indexer=self.reg, export=False,
                           info='Retrieved Qe of RenGen')

        self.Ipcmd = ExtAlgeb(model='RenGen', src='Ipcmd', indexer=self.reg, export=False,
                              info='Retrieved Ipcmd of RenGen',
                              e_str='-Ipcmd0 + IpHL_y',
                              )

        self.Iqcmd = ExtAlgeb(model='RenGen', src='Iqcmd', indexer=self.reg, export=False,
                              info='Retrieved Iqcmd of RenGen',
                              e_str='-Iqcmd0 - IqHL_y',   # negative sign here, different from `Ipcmd`
                              )

        self.p0 = ExtService(model='RenGen',
                             src='p0',
                             indexer=self.reg,
                             tex_name='P_0',
                             )
        self.q0 = ExtService(model='RenGen',
                             src='q0',
                             indexer=self.reg,
                             tex_name='Q_0',
                             )

        # Initial current commands
        self.Ipcmd0 = ConstService('p0 / v', info='initial Ipcmd')

        self.Iqcmd0 = ConstService('-q0 / v', info='initial Iqcmd')

        # --- Initial power factor angle ---
        # NOTE: if `p0` = 0, `pfaref0` = pi/2, `tan(pfaref0)` = inf
        self.pfaref0 = ConstService(v_str='atan2(q0, p0)', tex_name=r'\Phi_{ref0}',
                                    info='Initial power factor angle',
                                    )
        # flag devices with `p0`=0, which causes `tan(PF) = +inf`
        self.zp = IsEqual(self.p0, 0, cache=True)

        # --- Discrete components ---
        self.Vcmp = Limiter(u=self.v, lower=self.Vdip, upper=self.Vup, tex_name='V_{cmp}',
                            info='Voltage dip comparator', equal=False,
                            )
        self.Volt_dip = VarService(v_str='1 - Vcmp_zi',
                                   info='Voltage dip flag; 1-dip, 0-normal',
                                   tex_name='z_{Vdip}',
                                   )

        # --- Equations begin ---
        self.s0 = Lag(u=self.v, T=self.Trv, K=1,
                      info='Voltage filter',
                      )
        self.VLower = Limiter(u=self.v, lower=0.01, upper=999, no_upper=True,
                              info='Limiter for lower voltage cap',
                              )
        self.vp = Algeb(tex_name='V_p',
                        info='Sensed lower-capped voltage',
                        v_str='v * VLower_zi + 0.01 * VLower_zl',
                        e_str='v * VLower_zi + 0.01 * VLower_zl - vp',
                        )

        self.pfaref = Algeb(tex_name=r'\Phi_{ref}',
                            info='power factor angle ref',
                            unit='rad',
                            v_str='pfaref0',
                            e_str='pfaref0 - pfaref',
                            )

        self.S1 = Lag(u='Pe', T=self.Tp, K=1, tex_name='S_1', info='Pe filter',
                      )

        self.qref0 = ConstService(tex_name='q_{ref0}')
        self.qref0.v_str = 'SWQ_s1 * (v - Vref1) + SWQ_s0 * Iqcmd0 * (v * VLower_zi + 0.01 * VLower_zl)'
        self.Qref = Algeb(tex_name='Q_{ref}',
                          info='external Q ref',
                          v_str='-qref0',  # negative sign to be consistent with the direction of Qcpf
                          e_str='-qref0 - Qref',
                          unit='p.u.',
                          )

        # ignore `Qcpf` if `pfaref` is pi/2 by multiplying (1-zp_z1)
        self.Qcpf = Algeb(tex_name='Q_{cpf}',
                          info='Q calculated from P and power factor',
                          v_str='q0',
                          e_str='(1-zp_z1) * (S1_y * tan(pfaref) - Qcpf)',
                          diag_eps=True,
                          unit='p.u.',
                          )

        self.PFsel = Algeb(v_str='SWPF_s0*Qref + SWPF_s1*Qcpf',
                           e_str='SWPF_s0*Qref + SWPF_s1*Qcpf - PFsel',
                           info='Output of PFFLAG selector',
                           )

        self.PFlim = Limiter(u=self.PFsel, lower=self.QMin, upper=self.QMax)

        self.Qerr = Algeb(tex_name='Q_{err}',
                          info='Reactive power error',
                          v_str='(PFsel*PFlim_zi + QMin*PFlim_zl + QMax*PFlim_zu) - Qe',
                          e_str='(PFsel*PFlim_zi + QMin*PFlim_zl + QMax*PFlim_zu) - Qe - Qerr',
                          )

        self.PIQ = PITrackAWFreeze(u='SWV_s1 * Qerr + SWV_s0 * 0',
                                   kp=self.Kqp, ki=self.Kqi, ks=self.config.kqs,
                                   lower=self.VMIN, upper=self.VMAX,
                                   freeze=self.Volt_dip,
                                   )

        # If `VFLAG=0`, set the input as `Vref1` (see the NREL report)
        self.Vsel = GainLimiter(u='SWV_s0 * (Vref1 + SWPF_s0 * Qref + SWPF_s1 * Qcpf) + SWV_s1 * PIQ_y',
                                K=1, R=1,
                                lower=self.VMIN, upper=self.VMAX,
                                info='Selection output of VFLAG',
                                )

        # --- Placeholders for `Iqmin` and `Iqmax` ---
        self.s4 = LagFreeze(u='PFsel / vp',
                            T=self.Tiq, K=1,
                            freeze=self.Volt_dip,
                            tex_name='s_4',
                            info='Filter for calculated voltage with freeze',
                            )

        # --- Upper portion - Iqinj calculation ---

        self.Verr = Algeb(info='Voltage error (Vref0)',
                          v_str='Vref0 - s0_y',
                          e_str='Vref0 - s0_y - Verr',
                          tex_name='V_{err}',
                          )
        self.dbV = DeadBand1(u=self.Verr, lower=self.dbd1, upper=self.dbd2,
                             center=0.0,
                             enable='DB_{V}',
                             info='Deadband for voltage error (ref0)'
                             )

        self.pThld = ConstService(v_str='Indicator(Thld > 0)', tex_name='p_{Thld}')

        self.nThld = ConstService(v_str='Indicator(Thld < 0)', tex_name='n_{Thld}')

        self.Thld_abs = ConstService(v_str='abs(Thld)', tex_name='|Thld|')

        self.fThld = ExtendedEvent(self.Volt_dip,
                                   t_ext=self.Thld_abs,
                                   )

        # Gain after dbB
        Iqv = "(dbV_y * Kqv)"
        Iqinj = f'{Iqv} * Volt_dip + ' \
                f'(1 - Volt_dip) * fThld * ({Iqv} * nThld + Iqfrz * pThld)'

        # state transition, output of Iqinj
        self.Iqinj = Algeb(v_str=Iqinj,
                           e_str=Iqinj + ' - Iqinj',
                           tex_name='I_{qinj}',
                           info='Additional Iq signal during under- or over-voltage',
                           )

        # --- Lower portion - active power ---
        self.wg = Algeb(tex_name=r'\omega_g',
                        info='Drive train generator speed',
                        v_str='1.0',
                        e_str='1.0 - wg',
                        )

        self.Pref = Algeb(tex_name='P_{ref}',
                          info='external P ref',
                          v_str='p0 / wg',
                          e_str='p0 / wg - Pref',
                          unit='p.u.',
                          )

        self.pfilt = LagRate(u=self.Pref, T=self.config.tpfilt, K=1,
                             rate_lower=self.dPmin, rate_upper=self.dPmax,
                             info='Active power filter with rate limits',
                             tex_name='P_{filt}',
                             )

        self.Psel = Algeb(tex_name='P_{sel}',
                          info='Output selection of PFLAG',
                          v_str='SWP_s1*wg*pfilt_y + SWP_s0*pfilt_y',
                          e_str='SWP_s1*wg*pfilt_y + SWP_s0*pfilt_y - Psel',
                          )

        # `s5_y` is `Pord`
        self.s5 = LagAWFreeze(u=self.Psel, T=self.Tpord, K=1,
                              lower=self.PMIN, upper=self.PMAX,
                              freeze=self.Volt_dip,
                              tex_name='s5',
                              )

        self.Pord = AliasState(self.s5_y)

        # --- Current limit logic ---

        self.kVq12 = ConstService(v_str='(Iq2 - Iq1) / (Vq2 - Vq1)',
                                  tex_name='k_{Vq12}',
                                  )
        self.kVq23 = ConstService(v_str='(Iq3 - Iq2) / (Vq3 - Vq2)',
                                  tex_name='k_{Vq23}',
                                  )
        self.kVq34 = ConstService(v_str='(Iq4 - Iq3) / (Vq4 - Vq3)',
                                  tex_name='k_{Vq34}',
                                  )

        self.zVDL1 = ConstService(v_str='(Vq1 <= Vq2) & (Vq2 <= Vq3) & (Vq3 <= Vq4) & '
                                        '(Iq1 <= Iq2) & (Iq2 <= Iq3) & (Iq3 <= Iq4)',
                                  tex_name='z_{VDL1}',
                                  info='True if VDL1 is in service',
                                  )

        self.VDL1 = Piecewise(u=self.s0_y,
                              points=('Vq1', 'Vq2', 'Vq3', 'Vq4'),
                              funs=('Iq1',
                                    f'({self.s0_y.name} - Vq1) * kVq12 + Iq1',
                                    f'({self.s0_y.name} - Vq2) * kVq23 + Iq2',
                                    f'({self.s0_y.name} - Vq3) * kVq34 + Iq3',
                                    'Iq4'),
                              tex_name='V_{DL1}',
                              info='Piecewise linear characteristics of Vq-Iq',
                              )

        self.kVp12 = ConstService(v_str='(Ip2 - Ip1) / (Vp2 - Vp1)',
                                  tex_name='k_{Vp12}',
                                  )
        self.kVp23 = ConstService(v_str='(Ip3 - Ip2) / (Vp3 - Vp2)',
                                  tex_name='k_{Vp23}',
                                  )
        self.kVp34 = ConstService(v_str='(Ip4 - Ip3) / (Vp4 - Vp3)',
                                  tex_name='k_{Vp34}',
                                  )

        self.zVDL2 = ConstService(v_str='(Vp1 <= Vp2) & (Vp2 <= Vp3) & (Vp3 <= Vp4) & '
                                        '(Ip1 <= Ip2) & (Ip2 <= Ip3) & (Ip3 <= Ip4)',
                                  tex_name='z_{VDL2}',
                                  info='True if VDL2 is in service',
                                  )

        self.VDL2 = Piecewise(u=self.s0_y,
                              points=('Vp1', 'Vp2', 'Vp3', 'Vp4'),
                              funs=('Ip1',
                                    f'({self.s0_y.name} - Vp1) * kVp12 + Ip1',
                                    f'({self.s0_y.name} - Vp2) * kVp23 + Ip2',
                                    f'({self.s0_y.name} - Vp3) * kVp34 + Ip3',
                                    'Ip4'),
                              tex_name='V_{DL2}',
                              info='Piecewise linear characteristics of Vp-Ip',
                              )

        self.fThld2 = ExtendedEvent(self.Volt_dip,
                                    t_ext=self.Thld2,
                                    extend_only=True,
                                    )

        self.VDL1c = VarService(v_str='Lt(VDL1_y, Imaxr)')

        self.VDL2c = VarService(v_str='Lt(VDL2_y, Imaxr)')

        # `Iqmax` not considering mode or `Thld2`
        Iqmax1 = '(zVDL1*(VDL1c*VDL1_y + (1-VDL1c)*Imaxr) + 1e8*(1-zVDL1))'

        # `Ipmax` not considering mode or `Thld2`
        Ipmax1 = '(zVDL2*(VDL2c*VDL2_y + (1-VDL2c)*Imaxr) + 1e8*(1-zVDL2))'

        Ipmax2sq0 = '(Imax**2 - Iqcmd0**2)'

        Ipmax2sq = '(Imax**2 - IqHL_y**2)'

        # `Ipmax20`-squared (non-negative)
        self.Ipmax2sq0 = ConstService(v_str=f'Piecewise((0, Le({Ipmax2sq0}, 0.0)), ({Ipmax2sq0}, True), \
                                              evaluate=False)',
                                      tex_name='I_{pmax20,nn}^2',
                                      )

        self.Ipmax2sq = VarService(v_str=f'Piecewise((0, Le({Ipmax2sq}, 0.0)), ({Ipmax2sq}, True), \
                                           evaluate=False)',
                                   tex_name='I_{pmax2}^2',
                                   )

        Ipmax = f'((1-fThld2) * (SWPQ_s0*sqrt(Ipmax2sq) + SWPQ_s1*{Ipmax1}))'

        Ipmax0 = f'((1-fThld2) * (SWPQ_s0*sqrt(Ipmax2sq0) + SWPQ_s1*{Ipmax1}))'

        self.Ipmax = Algeb(v_str=f'{Ipmax0}',
                           e_str=f'{Ipmax} + (fThld2 * Ipmaxh) - Ipmax',
                           tex_name='I_{pmax}',
                           diag_eps=True,
                           info='Upper limit on Ipcmd',
                           )

        self.Ipmaxh = VarHold(self.Ipmax, hold=self.fThld2)

        Iqmax2sq = '(Imax**2 - IpHL_y**2)'

        Iqmax2sq0 = '(Imax**2 - Ipcmd0**2)'  # initialization equation by using `Ipcmd0`

        self.Iqmax2sq0 = ConstService(v_str=f'Piecewise((0, Le({Iqmax2sq0}, 0.0)), ({Iqmax2sq0}, True), \
                                              evaluate=False)',
                                      tex_name='I_{qmax,nn}^2',
                                      )

        self.Iqmax2sq = VarService(v_str=f'Piecewise((0, Le({Iqmax2sq}, 0.0)), ({Iqmax2sq}, True), \
                                           evaluate=False)',
                                   tex_name='I_{qmax2}^2')

        self.Iqmax = Algeb(v_str=f'(SWPQ_s0*{Iqmax1} + SWPQ_s1*sqrt(Iqmax2sq0))',
                           e_str=f'(SWPQ_s0*{Iqmax1} + SWPQ_s1*sqrt(Iqmax2sq)) - Iqmax',
                           tex_name='I_{qmax}',
                           info='Upper limit on Iqcmd',
                           )

        self.Iqmin = ApplyFunc(self.Iqmax, lambda x: -x, cache=False,
                               tex_name='I_{qmin}',
                               info='Lower limit on Iqcmd',
                               )

        self.Ipmin = ConstService(v_str='0.0', tex_name='I_{pmin}',
                                  info='Lower limit on Ipcmd',
                                  )

        self.PIV = PITrackAWFreeze(u='SWQ_s1 * (Vsel_y - s0_y * SWV_s0)',
                                   x0='-SWQ_s1 * Iqcmd0',
                                   kp=self.Kvp, ki=self.Kvi, ks=self.config.kvs,
                                   lower=self.Iqmin, upper=self.Iqmax,
                                   freeze=self.Volt_dip,
                                   )

        self.Qsel = Algeb(info='Selection output of QFLAG',
                          v_str='SWQ_s1 * PIV_y + SWQ_s0 * s4_y',
                          e_str='SWQ_s1 * PIV_y + SWQ_s0 * s4_y - Qsel',
                          tex_name='Q_{sel}',
                          )

        # `IpHL_y` is `Ipcmd`
        self.IpHL = GainLimiter(u='s5_y / vp',
                                K=1, R=1,
                                lower=self.Ipmin, upper=self.Ipmax,
                                )

        # `IqHL_y` is `Iqcmd`
        self.IqHL = GainLimiter(u='Qsel + Iqinj',
                                K=1, R=1,
                                lower=self.Iqmin, upper=self.Iqmax)


class REECA1(REECA1Data, REECA1Model):
    """
    Renewable energy electrical control.

    There are two user-defined voltages: `Vref0` and `Vref1`.

    - The difference between the initial bus voltage and `Vref0` should be
      within the voltage deadbands `dbd1` and `dbd2`.
    - If `VFLAG=0`, the input to the second PI controller will be `Vref1`.

    Regarding the additional reactive current injection during voltage dip:

    - Exercise caution when coordinating the parameters ``dbd1``, ``dbd2``,
      ``Vdip``, and ``Vup`` to avoid unintended responses.
    - ``Kqv`` in ``pu current / pu voltage deviation`` controls the intensity of
      reactive power injection. The parameter needs to tuned properly to avoid
      voltage overshoot.
    - When multiple renewable generators are connected to the same bus, ``Kqv``
      shall be reduced accordingly to avoid excessive reactive power injection.

    """

    def __init__(self, system, config):
        REECA1Data.__init__(self)
        REECA1Model.__init__(self, system, config)



from andes.models.exciter.ac8b import AC8B  # NOQA
from andes.models.exciter.esac1a import ESAC1A  # NOQA
from andes.models.exciter.esdc2a import ESDC2A  # NOQA
from andes.models.exciter.esst1a import ESST1A  # NOQA
from andes.models.exciter.esst3a import ESST3A  # NOQA
from andes.models.exciter.esst4b import ESST4B  # NOQA
from andes.models.exciter.exac1 import EXAC1  # NOQA
from andes.models.exciter.exac4 import EXAC4  # NOQA
from andes.models.exciter.exdc2 import EXDC2  # NOQA
from andes.models.exciter.exst1 import EXST1  # NOQA
from andes.models.exciter.ieeet1 import IEEET1  # NOQA
from andes.models.exciter.ieeet3 import IEEET3  # NOQA
from andes.models.exciter.ieeex1 import IEEEX1  # NOQA
from andes.models.exciter.saturation import ExcExpSat, ExcQuadSat  # NOQA
from andes.models.exciter.sexs import SEXS  # NOQA



"""
Switched shunt model.
"""

import ast
from collections import OrderedDict

import numpy as np

from andes.core import NumParam, ConstService
from andes.core.discrete import ShuntAdjust
from andes.core.service import SwBlock
from andes.models.shunt.shunt import ShuntData, ShuntModel


class ShuntSwData(ShuntData):
    """
    Data for switched shunts.
    """

    def __init__(self):
        ShuntData.__init__(self)
        self.gs = NumParam(info='a list literal of switched conductances blocks',
                           default=0.0,
                           unit='p.u.',
                           vtype=object,
                           iconvert=list_iconv,
                           oconvert=list_oconv,
                           y=True,
                           )

        self.bs = NumParam(info='a list literal of switched susceptances blocks',
                           default=0.0,
                           unit='p.u.',
                           vtype=object,
                           iconvert=list_iconv,
                           oconvert=list_oconv,
                           y=True,
                           )

        self.ns = NumParam(info='a list literal of the element numbers in each switched block',
                           default=[0],
                           vtype=object,
                           iconvert=list_iconv,
                           oconvert=list_oconv,
                           )

        self.vref = NumParam(info='voltage reference',
                             default=1.0,
                             unit='p.u.',
                             non_zero=True,
                             non_negative=True,
                             )

        self.dv = NumParam(info='voltage error deadband',
                           default=0.05,
                           unit='p.u.',
                           non_zero=True,
                           non_negative=True,
                           )

        self.dt = NumParam(info='delay before two consecutive switching',
                           default=30.,
                           unit='seconds',
                           non_negative=True,
                           )


def list_iconv(x):
    """
    Helper function to convert a list literal into a numpy array.
    """
    if isinstance(x, str):
        x = ast.literal_eval(x)
    if isinstance(x, (int, float)):
        if not np.isnan(x):
            x = [x]
        else:
            return None
    if isinstance(x, list):
        x = np.array(x)
    return x


def list_oconv(x):
    """
    Convert list into a list literal.
    """
    return np.array2string(x, separator=', ')


class ShuntSwModel(ShuntModel):
    """
    Switched shunt model.
    """

    def __init__(self, system, config):
        ShuntModel.__init__(self, system, config)

        self.config.add(OrderedDict((('min_iter', 2),
                                     ('err_tol', 0.01),
                                     )))
        self.config.add_extra("_help",
                              min_iter="iteration number starting from which to enable switching",
                              err_tol="iteration error below which to enable switching",
                              )
        self.config.add_extra("_alt",
                              min_iter='int',
                              err_tol='float',
                              )
        self.config.add_extra("_tex",
                              min_iter="sw_{iter}",
                              err_tol=r"\epsilon_{tol}",
                              )

        self.beff = SwBlock(init=self.b, ns=self.ns, blocks=self.bs)
        self.geff = SwBlock(init=self.g, ns=self.ns, blocks=self.gs,
                            ext_sel=self.beff)

        self.vlo = ConstService(v_str='vref - dv', tex_name='v_{lo}')
        self.vup = ConstService(v_str='vref + dv', tex_name='v_{up}')

        self.adj = ShuntAdjust(v=self.v, lower=self.vlo, upper=self.vup,
                               bsw=self.beff, gsw=self.geff, dt=self.dt,
                               u=self.u,
                               min_iter=self.config.min_iter,
                               err_tol=self.config.err_tol,
                               info='shunt adjuster')

        self.a.e_str = 'u * v**2 * geff'
        self.v.e_str = '-u * v**2 * beff'


class ShuntSw(ShuntSwData, ShuntSwModel):
    """
    Switched Shunt Model.

    Parameters `gs`, `bs` and `bs` must be entered in string literals,
    comma-separated. They need to have the same length.

    For example, in the excel file, one can put ::

        gs = [0, 0]
        bs = [0.2, 0.2]
        ns = [2, 4]

    To use individual shunts as fixed shunts, set the corresponding
    `ns = 0` or `ns = [0]`.

    The effective shunt susceptances and conductances are stored in
    services `beff` and `geff`.
    """

    def __init__(self, system=None, config=None):
        ShuntSwData.__init__(self)
        ShuntSwModel.__init__(self, system, config)



"""
Test PLBVFU1 model.
"""
import unittest
import andes


class TestPLBVFU1(unittest.TestCase):
    """
    Class for testing PLBVFU1.
    """

    def test_PLBVFU1(self):
        """
        Test PLVBFU1 model.
        """

        ss = andes.run(andes.get_case("ieee14/ieee14_plbvfu1.xlsx"),
                       default_config=True,
                       no_output=True,
                       )
        ss.TDS.config.tf = 3.0
        ss.TDS.config.criteria = 0
        ss.TDS.run()

        self.assertEqual(ss.exit_code, 0)



"""
Module for renewable energy generator (converter) model A with operator
splitting.
"""

from andes.core.service import VarService

from andes.models.renewable.regca1 import REGCA1Data, REGCA1Model


class REGCA1OSModel(REGCA1Model):
    """
    REGCA1OS implementation.
    """

    def __init__(self, system, config):
        REGCA1Model.__init__(self, system, config)
        delattr(self, 'LVG')
        delattr(self, 'LVG_y')
        delattr(self, 'Ipout')

        self.blocks.pop('LVG')
        self.algebs.pop('LVG_y')
        self.algebs.pop('Ipout')

        self.LVG_y = VarService(
            v_str='Piecewise((0, v <= Lvpnt0), \
                             ((v - Lvpnt0) * kLVG, v <= Lvpnt1), \
                             (1, v > Lvpnt1), \
                             (0, True))')
        self.Ipout = VarService(
            v_str='S0_y * LVG_y'
        )


class REGCA1OS(REGCA1Data, REGCA1OSModel):
    """
    Renewable energy generator model type A with operator splitting.
    """

    def __init__(self, system, config):
        REGCA1Data.__init__(self)
        REGCA1OSModel.__init__(self, system, config)



from andes.core import (Algeb, AntiWindup, ConstService, LagAntiWindup,
                        LeadLag, NumParam, State,)
from andes.core.block import DeadBand1
from andes.models.governor.tgbase import TGBase, TGBaseData


class TGOV1Data(TGBaseData):
    def __init__(self):
        super().__init__()
        self.R = NumParam(info='Speed regulation gain (mach. base default)',
                          tex_name='R',
                          default=0.05,
                          unit='p.u.',
                          ipower=True,
                          )
        self.VMAX = NumParam(info='Maximum valve position',
                             tex_name='V_{max}',
                             unit='p.u.',
                             default=1.2,
                             power=True,
                             )
        self.VMIN = NumParam(info='Minimum valve position',
                             tex_name='V_{min}',
                             unit='p.u.',
                             default=0.0,
                             power=True,
                             )

        self.T1 = NumParam(info='Valve time constant',
                           default=0.1,
                           tex_name='T_1')
        self.T2 = NumParam(info='Lead-lag lead time constant',
                           default=0.2,
                           tex_name='T_2')
        self.T3 = NumParam(info='Lead-lag lag time constant',
                           default=10.0,
                           tex_name='T_3')
        self.Dt = NumParam(info='Turbine damping coefficient',
                           default=0.0,
                           tex_name='D_t',
                           power=True,
                           )


class TGOV1DBData(TGOV1Data):
    def __init__(self):
        TGOV1Data.__init__(self)
        self.dbL = NumParam(info='Lower bound of deadband',
                            tex_name='db_L',
                            default=0.0,
                            unit='p.u.',
                            )
        self.dbU = NumParam(info='Upper bound of deadband',
                            tex_name='db_U',
                            default=0.0,
                            unit='p.u.',
                            )


class TGOV1Model(TGBase):
    """
    Implement TGOV1 model.
    """

    def __init__(self, system, config):
        TGBase.__init__(self, system, config)

        self.gain = ConstService(v_str='ue/R',
                                 tex_name='G',
                                 )

        self.pref = Algeb(info='Reference power input',
                          tex_name='P_{ref}',
                          v_str='tm0 * R',
                          e_str='pref0 * R - pref',
                          )

        self.wd = Algeb(info='Generator speed deviation',
                        unit='p.u.',
                        tex_name=r'\omega_{dev}',
                        v_str='0',
                        e_str='ue * (omega - wref) - wd',
                        )
        self.pd = Algeb(info='Pref plus speed deviation times gain',
                        unit='p.u.',
                        tex_name="P_d",
                        v_str='ue * tm0',
                        e_str='ue*(- wd + pref + paux) * gain - pd')

        self.LAG = LagAntiWindup(u=self.pd,
                                 K=1,
                                 T=self.T1,
                                 lower=self.VMIN,
                                 upper=self.VMAX,
                                 )
        self.LL = LeadLag(u=self.LAG_y,
                          T1=self.T2,
                          T2=self.T3,
                          )
        self.pout.e_str = 'ue * (LL_y - Dt * wd) - pout'


class TGOV1NModel(TGOV1Model):
    """
    New TGOV1 model with `pref` and `paux` summed after the gain.
    """

    def __init__(self, system, config):
        TGOV1Model.__init__(self, system, config)
        self.pref.v_str = 'tm0'
        self.pref.e_str = 'pref0 - pref'

        self.pd.e_str = 'ue*(-wd * gain + pref + paux) - pd'


class TGOV1DBModel(TGOV1Model):
    """
    Model TGOV1 with deadband.
    """

    def __init__(self, system, config):
        TGOV1Model.__init__(self, system, config)
        self.DB = DeadBand1(u=self.wd, center=0.0, lower=self.dbL,
                            upper=self.dbU, tex_name='DB',
                            info='deadband for speed deviation',
                            )
        self.pd.e_str = 'ue * (-DB_y + pref + paux) * gain - pd'
        self.pout.e_str = '(LL_y - Dt * DB_y) - pout'


class TGOV1NDBModel(TGOV1DBModel):
    """
    Implementation of TGOV1NDB
    """

    def __init__(self, system, config):
        TGOV1DBModel.__init__(self, system, config)
        self.pref.v_str = 'tm0'
        self.pref.e_str = 'pref0 - pref'

        self.pd.e_str = 'ue*(DB_y * gain + pref + paux) - pd'


class TGOV1ModelAlt(TGBase):
    """
    An alternative implementation of TGOV1 from equations
    (without using Blocks).
    """

    def __init__(self, system, config):
        TGBase.__init__(self, system, config)

        self.pref = Algeb(info='Reference power input',
                          tex_name='P_{ref}',
                          v_str='tm0 * R',
                          e_str='pref0 * R - pref',
                          )
        self.wd = Algeb(info='Generator speed deviation',
                        unit='p.u.',
                        tex_name=r'\omega_{dev}',
                        v_str='0',
                        e_str='ue * (omega - wref) - wd',
                        )
        self.pd = Algeb(info='Pref plus speed deviation times gain',
                        unit='p.u.',
                        tex_name="P_d",
                        v_str='tm0',
                        e_str='(- wd + pref + paux) * gain - pd')

        self.LAG_y = State(info='State in lag transfer function',
                           tex_name=r"x'_{LAG}",
                           e_str='LAG_lim_zi * (1 * pd - LAG_y)',
                           t_const=self.T1,
                           v_str='pd',
                           )
        self.LAG_lim = AntiWindup(u=self.LAG_y,
                                  lower=self.VMIN,
                                  upper=self.VMAX,
                                  tex_name='lim_{lag}',
                                  )
        self.LL_x = State(info='State in lead-lag transfer function',
                          tex_name="x'_{LL}",
                          v_str='LAG_y',
                          e_str='(LAG_y - LL_x)',
                          t_const=self.T3
                          )
        self.LL_y = Algeb(info='Lead-lag Output',
                          tex_name='y_{LL}',
                          v_str='LAG_y',
                          e_str='T2 / T3 * (LAG_y - LL_x) + LL_x - LL_y',
                          )

        self.pout.e_str = 'ue * (LL_y - Dt * wd) - pout'


class TGOV1(TGOV1Data, TGOV1Model):
    """
    TGOV1 turbine governor model.

    Implements the PSS/E TGOV1 model without deadband.
    """

    def __init__(self, system, config):
        TGOV1Data.__init__(self)
        TGOV1Model.__init__(self, system, config)


class TGOV1N(TGOV1Data, TGOV1NModel):
    """
    New TGOV1 (TGOV1N) turbine governor model.

    New TGOV1 model with `pref` and `paux` summed after the gain.
    This model is useful for incorporating AGC and scheduling
    signals without having to know the droop.

    Scheduling changes should write to the `v` fields of
    `pref0` and `qref0` in place.
    AGC signal should write to that of `paux0` in place.

    Modifying `tm0` is not allowed.
    """

    def __init__(self, system, config):
        TGOV1Data.__init__(self)
        TGOV1NModel.__init__(self, system, config)


class TGOV1DB(TGOV1DBData, TGOV1DBModel):
    """
    TGOV1 turbine governor model with speed input deadband.
    """

    def __init__(self, system, config):
        TGOV1DBData.__init__(self)
        TGOV1DBModel.__init__(self, system, config)


class TGOV1NDB(TGOV1DBData, TGOV1NDBModel):
    """
    TGOV1N turbine governor model with speed input deadband.
    """

    def __init__(self, system, config):
        TGOV1DBData.__init__(self)
        TGOV1NDBModel.__init__(self, system, config)



"""
Third-order motor model.
"""

from andes.models.motor.motorbase import MotorBaseModel, MotorBaseData


class Motor3Model(MotorBaseModel):
    """
    Third-order induction motor model
    """

    def __init__(self, system, config):
        MotorBaseModel.__init__(self, system, config)

        self.Id.e_str = 'u * (vd - e1d - rs * Id + x1 * Iq)'

        self.Id.v_str = '1'

        self.Iq.e_str = 'u * (vq - e1q - rs * Iq - x1 * Id)'

        self.te.v_str = 'u * (e1d * Id + e1q * Iq)'

        self.te.e_str = f'{self.te.v_str} - te'


class Motor3(MotorBaseData, Motor3Model):
    """
    Third-order induction motor model.

    See "Power System Modelling and Scripting" by F. Milano.

    To simulate motor startup, set the motor status ``u`` to ``0``
    and use a ``Toggler`` to control the model.
    """

    def __init__(self, system, config):
        MotorBaseData.__init__(self)
        Motor3Model.__init__(self, system, config)



from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        IdxParam, Model, ModelData, NumParam, State,)
from andes.core.block import Integrator
from andes.core.var import AliasState


class WTDSData(ModelData):
    """
    Wind turbine governor swing equation model data.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.ree = IdxParam(mandatory=True,
                            info='Renewable exciter idx',
                            )

        self.H = NumParam(default=3.0, tex_name='H_t',
                          info='Total inertia', unit='MWs/MVA',
                          power=True,
                          non_zero=True,
                          non_negative=True,
                          )

        self.D = NumParam(default=1.0, tex_name='D_{shaft}',
                          info='Damping coefficient',
                          unit='p.u.',
                          power=True,
                          )

        self.w0 = NumParam(default=1.0, tex_name=r'\omega_0',
                           info='Default speed if not using a torque model',
                           unit='p.u.',
                           )


class WTDSModel(Model):
    """
    Wind turbine one-mass generator model.

    User-provided reference speed should be specified in parameter `w0`.
    Internally, `w0` is set to the algebraic variable `wr0`.

    See J.H. Chow, J.J. Sanchez-Gasca. "Power System Modeling, Computation, and Control".
    John Wiley & Sons, 2020. pp. 518.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.flags.tds = True
        self.group = 'RenGovernor'

        self.reg = ExtParam(model='RenExciter', src='reg', indexer=self.ree,
                            export=False,
                            )

        self.Sn = ExtParam(model='RenGen', src='Sn', indexer=self.reg,
                           tex_name='S_n', export=False,
                           )

        self.wge = ExtAlgeb(model='RenExciter', src='wg', indexer=self.ree,
                            export=False,
                            e_str='-1.0 + s1_y',
                            ename='wg',
                            tex_ename=r'\omega_g',
                            )

        self.Pe = ExtAlgeb(model='RenGen', src='Pe', indexer=self.reg, export=False,
                           info='Retrieved Pe of RenGen')

        self.Pe0 = ExtService(model='RenGen', src='Pe', indexer=self.reg, tex_name='P_{e0}',
                              )

        self.H2 = ConstService(v_str='2 * H', tex_name='2H')

        self.Pm = Algeb(tex_name='P_m',
                        info='Mechanical power',
                        e_str='Pe0 - Pm',
                        v_str='Pe0',
                        )

        self.wr0 = Algeb(tex_name=r'\omega_{r0}',
                         unit='p.u.',
                         v_str='w0',
                         e_str='w0 - wr0',
                         info='speed set point',
                         )

        # `s1_y` is `w_m`
        self.s1 = Integrator(u='(Pm - Pe) / wge - D * (s1_y - wr0)',
                             T=self.H2,
                             K=1.0,
                             y0='wr0',
                             )

        # make two alias states, `wt` and `wg`, pointing to `s1_y`
        self.wt = AliasState(self.s1_y, tex_name=r'\omega_t')

        self.wg = AliasState(self.s1_y, tex_name=r'\omega_g')

        self.s3_y = State(info='Unused state variable',
                          tex_name='y_{s3}',
                          )

        self.Kshaft = ConstService(v_str='1.0', tex_name='K_{shaft}',
                                   info='Dummy Kshaft',
                                   )


class WTDS(WTDSData, WTDSModel):
    """
    Custom wind turbine model with a single swing-equation.

    This model is used to simulate the mechanical swing
    of the combined machine and turbine mass. The speed output
    is ``s1_y`` which will be fed to ``RenExciter.wg``.

    ``PFLAG`` needs to be set to ``1`` in exciter to consider
    speed for Pref.
    """

    def __init__(self, system, config):
        WTDSData.__init__(self)
        WTDSModel.__init__(self, system, config)



from andes.core import ModelData, IdxParam, NumParam, Model, ConstService, ExtAlgeb, Algeb, State


class MotorBaseData(ModelData):
    """Base parameters for induction machines
    """

    def __init__(self):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="interface bus id",
                            mandatory=True,
                            )

        self.Sn = NumParam(default=100.0,
                           info="Power rating",
                           tex_name='S_n',
                           )

        self.Vn = NumParam(default=110.0,
                           info="AC voltage rating",
                           tex_name='V_n',
                           )
        self.fn = NumParam(default=60.0,
                           info="rated frequency",
                           tex_name='f',
                           )

        self.rs = NumParam(default=0.01,
                           info="rotor resistance",
                           z=True,
                           tex_name='r_s',
                           non_zero=True,
                           )

        self.xs = NumParam(default=0.15,
                           info="rotor reactance",
                           z=True,
                           tex_name='x_s',
                           non_zero=True,
                           )
        self.rr1 = NumParam(default=0.05,
                            info="1st cage rotor resistance",
                            z=True,
                            non_zero=True,
                            tex_name='r_{R1}',
                            )
        self.xr1 = NumParam(default=0.15,
                            info="1st cage rotor reactance",
                            z=True,
                            tex_name='x_{R1}',
                            non_zero=True,
                            )
        self.rr2 = NumParam(default=0.001,
                            info="2st cage rotor resistance",
                            z=True,
                            non_zero=True,
                            tex_name='r_{R2}',
                            )
        self.xr2 = NumParam(default=0.04,
                            info="2st cage rotor reactance",
                            z=True,
                            tex_name='x_{R2}',
                            non_zero=True,
                            )

        self.xm = NumParam(default=5.0,
                           info="magnetization reactance",
                           z=True,
                           tex_name='x_m',
                           non_zero=True,
                           )

        self.Hm = NumParam(default=3.0,
                           info='Inertia constant',
                           power=True,
                           tex_name='H_m',
                           unit='kWs/KVA',
                           )

        self.c1 = NumParam(default=0.1,
                           info='1st coeff. of Tm(w)',
                           tex_name='c_1',
                           )

        self.c2 = NumParam(default=0.02,
                           info='2nd coeff. of Tm(w)',
                           tex_name='c_2',
                           )

        self.c3 = NumParam(default=0.02,
                           info='3rd coeff. of Tm(w)',
                           tex_name='c_3',
                           )

        self.zb = NumParam(default=1.0,
                           info='Allow working as brake',
                           tex_name='z_b',
                           vrange=(0, 1),
                           )


class MotorBaseModel(Model):
    """Base model for motors
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)

        self.flags.pflow = True
        self.flags.tds = True
        self.flags.tds_init = False
        self.group = 'Motor'

        # services
        self.wb = ConstService(v_str='2 * pi * fn',
                               tex_name=r'\omega_b',
                               )

        self.x0 = ConstService(v_str='xs + xm',
                               tex_name='x_0',
                               )

        self.x1 = ConstService(v_str='xs + xr1 * xm / (xr1 + xm)',
                               tex_name="x'",
                               )

        self.T10 = ConstService(v_str='(xr1 + xm) / (wb * rr1)',
                                tex_name="T'_0",
                                )

        self.M = ConstService(v_str='2 * Hm',
                              tex_name='M',
                              )

        self.aa = ConstService(v_str='c1 + c2 + c3',
                               tex_name=r'\alpha',
                               )

        self.bb = ConstService(v_str='-c2 - 2 * c3',
                               tex_name=r'\beta',
                               )

        # network algebraic variables
        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage phase angle',
                          e_str='+p',
                          ename='P',
                          tex_ename='P',
                          )

        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          info='Bus voltage magnitude',
                          e_str='+q',
                          ename='Q',
                          tex_ename='Q',
                          )

        self.vd = Algeb(info='d-axis voltage',
                        e_str='-u * v * sin(a) - vd',
                        tex_name=r'V_d',
                        )

        self.vq = Algeb(info='q-axis voltage',
                        e_str='u * v * cos(a) - vq',
                        tex_name=r'V_q',
                        )

        self.slip = State(tex_name=r"\sigma",
                          e_str='u * (tm - te)',
                          t_const=self.M,
                          diag_eps=True,
                          v_str='1.0 * u',
                          )

        self.p = Algeb(tex_name='P',
                       e_str='u * (vd * Id + vq * Iq) - p',
                       v_str='u * (vd * Id + vq * Iq)',
                       )

        self.q = Algeb(tex_name='Q',
                       e_str='u * (vq * Id - vd * Iq) - q',
                       v_str='u * (vq * Id - vd * Iq)',
                       )

        self.e1d = State(info='real part of 1st cage voltage',
                         tex_name="e'_d",
                         v_str='0.05 * u',
                         e_str='u * (wb*slip*e1q - (e1d + (x0 - x1) * Iq)/T10)',
                         diag_eps=True,
                         )

        self.e1q = State(info='imaginary part of 1st cage voltage',
                         tex_name="e'_q",
                         v_str='0.9 * u',
                         e_str='u * (-wb*slip*e1d - (e1q - (x0 - x1) * Id)/T10)',
                         diag_eps=True,
                         )

        self.Id = Algeb(tex_name='I_d',
                        diag_eps=True,
                        )

        self.Iq = Algeb(tex_name='I_q',
                        diag_eps=True,
                        )

        self.te = Algeb(tex_name=r'\tau_e',
                        )

        self.tm = Algeb(tex_name=r'\tau_m',
                        )

        self.tm.v_str = 'u * (aa + bb * slip + c2 * slip * slip)'
        self.tm.e_str = f'{self.tm.v_str} - tm'



"""
Package for synchronous generators
"""

from andes.models.synchronous.gencls import GENCLS  # NOQA
from andes.models.synchronous.genrou import GENROU  # NOQA
from andes.models.synchronous.plbvfu1 import PLBVFU1  # NOQA



"""
Steady-state PV model.
"""

from collections import OrderedDict

from andes.core import (ModelData, NumParam, ExtParam, DataParam, IdxParam, Model,
                        BackRef, ExtAlgeb, Algeb, SortedLimiter)
from andes.core.service import ConstService


class PVData(ModelData):
    def __init__(self):
        super().__init__()
        self.Sn = NumParam(default=100.0, info="Power rating", non_zero=True, tex_name=r'S_n')
        self.Vn = NumParam(default=110.0, info="AC voltage rating", non_zero=True, tex_name=r'V_n')
        self.subidx = DataParam(info='index for generators on the same bus', export=False)
        self.bus = IdxParam(model='Bus', info="idx of the installed bus", mandatory=True)
        self.busr = IdxParam(model='Bus', info="bus idx for remote voltage control")
        self.p0 = NumParam(default=0.0, info="active power set point in system base", tex_name=r'p_0', unit='p.u.')
        self.q0 = NumParam(default=0.0, info="reactive power set point in system base", tex_name=r'q_0',
                           unit='p.u.')

        self.pmax = NumParam(default=999.0, info="maximum active power in system base",
                             tex_name=r'p_{max}', unit='p.u.')
        self.pmin = NumParam(default=-1.0, info="minimum active power in system base",
                             tex_name=r'p_{min}', unit='p.u.')
        self.qmax = NumParam(default=999.0, info="maximim reactive power in system base",
                             tex_name=r'q_{max}', unit='p.u.')
        self.qmin = NumParam(default=-999.0, info="minimum reactive power in system base",
                             tex_name=r'q_{min}', unit='p.u.')

        self.v0 = NumParam(default=1.0, info="voltage set point", tex_name=r'v_0')
        self.vmax = NumParam(default=1.4, info="maximum voltage voltage", tex_name=r'v_{max}')
        self.vmin = NumParam(default=0.6, info="minimum allowed voltage", tex_name=r'v_{min}')
        self.ra = NumParam(default=0.0, info='armature resistance', tex_name='r_a')
        self.xs = NumParam(default=0.3, info='armature reactance', tex_name='x_s')


class PVModel(Model):
    """
    PV generator model (power flow) with q limit and PV-PQ conversion.
    """

    def __init__(self, system=None, config=None):
        super().__init__(system, config)
        self.group = 'StaticGen'
        self.flags.pflow = True
        self.flags.tds = True
        self.flags.tds_init = False

        self.config.add(OrderedDict((('pv2pq', 0),
                                     ('npv2pq', 0),
                                     ('min_iter', 2),
                                     ('err_tol', 0.01),
                                     ('abs_violation', 1),
                                     )))
        self.config.add_extra("_help",
                              pv2pq="convert PV to PQ in PFlow at Q limits",
                              npv2pq="max. # of conversion each iteration, 0 - auto",
                              min_iter="iteration number starting from which to enable switching",
                              err_tol="iteration error below which to enable switching",
                              abs_violation='use absolute (1) or relative (0) limit violation',
                              )

        self.config.add_extra("_alt",
                              pv2pq=(0, 1),
                              npv2pq=">=0",
                              min_iter='int',
                              err_tol='float',
                              abs_violation=(0, 1),
                              )
        self.config.add_extra("_tex",
                              pv2pq="z_{pv2pq}",
                              npv2pq="n_{pv2pq}",
                              min_iter="sw_{iter}",
                              err_tol=r"\epsilon_{tol}"
                              )

        self.SynGen = BackRef()

        self.busv0 = ExtParam(model='Bus', src='v0', indexer=self.bus,
                              export=False, tex_name='V_{0bus}',
                              )

        self.a = ExtAlgeb(model='Bus', src='a', indexer=self.bus, tex_name=r'\theta',
                          ename='P',
                          tex_ename='P',
                          is_input=True,
                          )
        self.v = ExtAlgeb(model='Bus', src='v', indexer=self.bus, v_setter=True, tex_name=r'V',
                          ename='dV',
                          tex_ename=r'\Delta V',
                          is_input=True,
                          )

        self.p = ConstService(v_str='p0',
                              info='copy of p0 used for power flow',
                              tex_name='p',)

        self.q = Algeb(info='actual reactive power generation',
                       unit='p.u.',
                       tex_name='q',
                       diag_eps=True,
                       )

        self.qlim = SortedLimiter(u=self.q, lower=self.qmin, upper=self.qmax,
                                  enable=self.config.pv2pq,
                                  n_select=self.config.npv2pq,
                                  min_iter=self.config.min_iter,
                                  err_tol=self.config.err_tol,
                                  abs_violation=self.config.abs_violation,
                                  )

        # variable initialization equations
        self.v.v_str = 'u * v0 + (1-u) * busv0'
        self.q.v_str = 'u * q0'

        # injections into buses have negative values
        self.a.e_str = "-u * p"
        self.v.e_str = "-u * q"

        # power injection equations g(y) = 0
        self.q.e_str = "u*(qlim_zi * (v0-v) + " \
                       "qlim_zl * (qmin-q) + " \
                       "qlim_zu * (qmax-q))"

        # NOTE: the line below fails at TDS
        # self.q.e_str = "Piecewise((qmin - q, q < qmin), (qmax - q, q > qmax), (v0 - v, True))"


class PV(PVData, PVModel):
    """
    Static PV generator with reactive power limit checking
    and PV-to-PQ conversion.

    `pv2pq = 1` turns on the conversion.
    It starts  from iteration `min_iter` or when the convergence
    error drops below `err_tol`.

    The PV-to-PQ conversion first ranks the reactive violations.
    A maximum number of `npv2pq` PVs above the upper limit, and
    a maximum of `npv2pq` PVs below the lower limit will be
    converted to PQ, which sets the reactive power to `pmax` or
    `pmin`.

    If `pv2pq` is `1` (enabled) and `npv2pq` is `0`, heuristics
    will be used to determine the number of PVs to be converted
    for each iteration.
    """

    def __init__(self, system=None, config=None):
        PVData.__init__(self)
        PVModel.__init__(self, system, config)



"""
Bus rate-of-change of frequency measurement based on BusFreq.
"""

from andes.core import NumParam, Washout
from andes.models.measurement import BusFreq


class BusROCOF(BusFreq):
    """
    Bus frequency and ROCOF measurement.

    The ROCOF output variable is ``Wf_y``.
    """

    def __init__(self, system, config):
        BusFreq.__init__(self, system, config)
        self.Tr = NumParam(default=0.1,
                           info="frequency washout time constant",
                           tex_name='T_r')

        self.Wf = Washout(u=self.f,
                          K=1,
                          T=self.Tr,
                          info='frequency washout yielding ROCOF',
                          )



"""
ESST1A model.
"""

from andes.core.block import (GainLimiter, HVGate, Lag, LagAntiWindup, LeadLag,
                              LVGate, Washout,)
from andes.core.discrete import Switcher
from andes.core.param import NumParam
from andes.core.service import ConstService, PostInitService, VarService
from andes.core.var import Algeb, ExtAlgeb
from andes.models.exciter.excbase import ExcBase, ExcBaseData, ExcVsum


class ESST1AData(ExcBaseData):
    """
    ESST1A data.
    """

    def __init__(self):
        ExcBaseData.__init__(self)
        self.TR = NumParam(info='Sensing time constant',
                           tex_name='T_R',
                           default=0.01,
                           )

        self.VIMAX = NumParam(default=0.8,
                              info='Max. input voltage',
                              tex_name=r'V_{IMAX}',
                              )
        self.VIMIN = NumParam(default=-0.1,
                              info='Min. input voltage',
                              tex_name=r'V_{IMIN}',
                              )
        self.TB = NumParam(info='Lag time constant in lead-lag',
                           tex_name='T_B',
                           default=1,
                           )
        self.TC = NumParam(info='Lead time constant in lead-lag',
                           tex_name='T_C',
                           default=1,
                           )

        self.TB1 = NumParam(info='Lag time constant in lead-lag 1',
                            tex_name=r'T_{B1}',
                            default=1,
                            )
        self.TC1 = NumParam(info='Lead time constant in lead-lag 1',
                            tex_name=r'T_{C1}',
                            default=1,
                            )

        self.VAMAX = NumParam(info='V_A upper limit',
                              tex_name=r'V_{AMAX}',
                              default=999,
                              unit='p.u.')
        self.VAMIN = NumParam(info='V_A lower limit',
                              tex_name=r'V_{AMIN}',
                              default=-999,
                              unit='p.u.')

        self.KA = NumParam(default=80,
                           info='Regulator gain',
                           tex_name='K_A',
                           )
        self.TA = NumParam(info='Lag time constant in regulator',
                           tex_name='T_A',
                           default=0.04,
                           )

        self.ILR = NumParam(default=1,
                            info='Exciter output current limite reference',
                            tex_name=r'I_{LR}',
                            )
        self.KLR = NumParam(default=1,
                            info='Exciter output current limiter gain',
                            tex_name=r'K_{LR}',
                            )

        self.VRMAX = NumParam(info='Maximum voltage regulator output limit',
                              tex_name=r'V_{RMAX}',
                              default=7.3,
                              unit='p.u.',)
        self.VRMIN = NumParam(info='Minimum voltage regulator output limit',
                              tex_name=r'V_{RMIN}',
                              default=-7.3,
                              unit='p.u.',)

        self.KF = NumParam(default=0.1,
                           info='Feedback gain',
                           tex_name='K_F',
                           )
        self.TF = NumParam(info='Feedback washout time constant',
                           tex_name='T_{F}',
                           default=1,
                           )

        self.KC = NumParam(info='Rectifier loading factor proportional to commutating reactance',
                           tex_name='K_C',
                           default=0.1,
                           )

        self.UELc = NumParam(info='Alternate UEL inputs, input code 1-3',
                             tex_name='UEL_c',
                             default=1,
                             )
        self.VOSc = NumParam(info='Alternate Stabilizer inputs, input code 1-2',
                             tex_name='VOS_c',
                             default=1,
                             )


class ESST1AModel(ExcBase, ExcVsum):
    """
    Implementation of the ESST1A model.
    """

    def __init__(self, system, config):
        ExcBase.__init__(self, system, config)
        self.flags.nr_iter = True

        ExcVsum.__init__(self)
        self.UEL0.v_str = '-999'
        self.OEL0.v_str = '999'

        self.ulim = ConstService('9999')
        self.llim = ConstService('-9999')

        self.SWUEL = Switcher(u=self.UELc, options=[0, 1, 2, 3], tex_name='SW_{UEL}', cache=True)
        self.SWVOS = Switcher(u=self.VOSc, options=[0, 1, 2], tex_name='SW_{VOS}', cache=True)

        # control block begin
        self.LG = Lag(self.v, T=self.TR, K=1,
                      info='Voltage transducer',
                      )
        self.SG0 = ConstService(v_str='0', info='SG initial value.')
        self.SG = Algeb(tex_name='SG', info='SG',
                        v_str='SG0',
                        e_str='SG0 - SG',
                        )

        self.zero = ConstService('0')
        self.LR = GainLimiter(u='XadIfd - ILR',
                              K=self.KLR, R=1,
                              upper=self.ulim, lower=self.zero,
                              no_upper=True,
                              info='Exciter output current gain limiter',
                              )

        self.VA0 = PostInitService(tex_name='V_{A0}',
                                   v_str='vf0 - SWVOS_s2 * SG + LR_y',
                                   info='VA (LA_y) initial value')

        self.vref.v_str = 'ue * (v + (vf0 - SWVOS_s2 * SG + LR_y) / KA - SWVOS_s1 * SG - SWUEL_s1 * UEL)'
        self.vref.v_iter = 'ue * (v + (vf0 - SWVOS_s2 * SG + LR_y) / KA - SWVOS_s1 * SG - SWUEL_s1 * UEL)'

        self.vref0 = PostInitService(info='Initial reference voltage input',
                                     tex_name='V_{ref0}',
                                     v_str='vref',
                                     )

        self.vi = Algeb(info='Total input voltages',
                        tex_name='V_i',
                        unit='p.u.',
                        e_str='ue * (-LG_y + vref - WF_y + SWUEL_s1 * UEL + SWVOS_s1 * SG + Vs) - vi',
                        v_iter='ue * (-LG_y + vref - WF_y + SWUEL_s1 * UEL + SWVOS_s1 * SG + Vs)',
                        v_str='ue * (-LG_y + vref - WF_y + SWUEL_s1 * UEL + SWVOS_s1 * SG + Vs)',
                        )

        self.vil = GainLimiter(u=self.vi,
                               K=1, R=1,
                               upper=self.VIMAX, lower=self.VIMIN,
                               info='Exciter voltage input limiter',
                               )

        self.UEL2 = Algeb(tex_name='UEL_2',
                          info='UEL_2 as HVG1 u1',
                          v_str='ue * (SWUEL_s2 * UEL + (1 - SWUEL_s2) * llim)',
                          e_str='ue * (SWUEL_s2 * UEL + (1 - SWUEL_s2) * llim) - UEL2',
                          )
        self.HVG1 = HVGate(u1=self.UEL2,
                           u2=self.vil_y,
                           info='HVGate after V_I',
                           )

        self.LL = LeadLag(u=self.HVG1_y,
                          T1=self.TC,
                          T2=self.TB,
                          info='Lead-lag compensator',
                          zero_out=True,
                          )

        self.LL1 = LeadLag(u=self.LL_y,
                           T1=self.TC1,
                           T2=self.TB1,
                           info='Lead-lag compensator 1',
                           zero_out=True,
                           )

        self.LA = LagAntiWindup(u=self.LL1_y,
                                T=self.TA,
                                K=self.KA,
                                upper=self.VAMAX,
                                lower=self.VAMIN,
                                info='V_A, Anti-windup lag',
                                )  # LA_y is VA

        self.vas = Algeb(tex_name=r'V_{As}',
                         info='V_A after subtraction, as HVG u2',
                         v_str='ue * (SWVOS_s2 * SG + LA_y - LR_y)',
                         v_iter='ue * (SWVOS_s2 * SG + LA_y - LR_y)',
                         e_str='ue * (SWVOS_s2 * SG + LA_y - LR_y) - vas',
                         )

        self.UEL3 = Algeb(tex_name='UEL_3',
                          info='UEL_3 as HVG u1',
                          v_str='ue * (SWUEL_s3 * UEL + (1 - SWUEL_s3) * llim)',
                          e_str='ue * (SWUEL_s3 * UEL + (1 - SWUEL_s3) * llim) - UEL3',
                          )
        self.HVG = HVGate(u1=self.UEL3,
                          u2=self.vas,
                          info='HVGate for under excitation',
                          )

        self.LVG = LVGate(u1=self.HVG_y,
                          u2=self.OEL,
                          info='HVGate for over excitation',
                          )

        # vd, vq, Id, Iq from SynGen
        self.vd = ExtAlgeb(src='vd',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'V_d',
                           info='d-axis machine voltage',
                           )
        self.vq = ExtAlgeb(src='vq',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'V_q',
                           info='q-axis machine voltage',
                           )

        self.efdu = VarService(info='Output exciter voltage upper bound',
                               tex_name=r'efd_{u}',
                               v_str='Abs(vd + 1j*vq) * VRMAX - KC * XadIfd',
                               sequential=False,
                               )
        self.efdl = VarService(info='Output exciter voltage lower bound',
                               tex_name=r'efd_{l}',
                               v_str='Abs(vd + 1j*vq) * VRMIN',
                               sequential=False,
                               )

        self.vol = GainLimiter(u=self.LVG_y,
                               K=1, R=1,
                               upper=self.efdu,
                               lower=self.efdl,
                               info='Exciter output limiter',
                               )

        self.WF = Washout(u=self.LVG_y,
                          T=self.TF,
                          K=self.KF,
                          info='V_F, Stablizing circuit feedback',
                          )

        self.vout.e_str = 'ue * vol_y  - vout'


class ESST1A(ESST1AData, ESST1AModel):
    """
    Exciter ESST1A model.

    Reference:

    [1] PowerWorld, Exciter ESST1A, [Online],

    [2] NEPLAN, Exciters Models, [Online],

    Available:
    https://www.powerworld.com/WebHelp/Content/TransientModels_HTML/Exciter%20ESST1A.htm

    https://www.neplan.ch/wp-content/uploads/2015/08/Nep_EXCITERS1.pdf
    """

    def __init__(self, system, config):
        ESST1AData.__init__(self)
        ESST1AModel.__init__(self, system, config)



from andes.core.block import Integrator, Lag, LagAntiWindup, LessThan, Washout
from andes.core.param import NumParam
from andes.core.service import ConstService, FlagValue, PostInitService
from andes.core.var import Algeb
from andes.models.exciter.excbase import ExcBase, ExcBaseData
from andes.models.exciter.saturation import ExcQuadSat


class IEEET1Data(ExcBaseData):
    def __init__(self):
        ExcBaseData.__init__(self)

        self.TR = NumParam(info='Sensing time constant',
                           tex_name='T_R',
                           default=0.02,
                           unit='p.u.',
                           )
        self.KA = NumParam(info='Regulator gain',
                           tex_name='K_A',
                           default=5.0,
                           unit='p.u.',
                           )
        self.TA = NumParam(info='Lag time constant in anti-windup lag',
                           tex_name='T_A',
                           default=0.04,
                           unit='p.u.',
                           )
        self.VRMAX = NumParam(info='Maximum excitation limit',
                              tex_name='V_{RMAX}',
                              default=7.3,
                              unit='p.u.')
        self.VRMIN = NumParam(info='Minimum excitation limit',
                              tex_name='V_{RMIN}',
                              default=-7.3,
                              unit='p.u.')
        self.KE = NumParam(info='Gain added to saturation',
                           tex_name='K_E',
                           default=1,
                           unit='p.u.',
                           )
        self.TE = NumParam(info='Exciter integrator time constant',
                           tex_name='T_E',
                           default=0.8,
                           unit='p.u.',
                           )
        self.KF = NumParam(default=0.1,
                           info='Feedback gain',
                           tex_name='K_F',
                           )
        self.TF = NumParam(default=1.0,
                           info='Feedback delay',
                           tex_name='T_F',
                           non_negative=True,
                           non_zero=True,
                           )
        self.Switch = NumParam(info='Switch unused in PSS/E',
                               tex_name='S_w',
                               default=0,
                               unit='bool',
                               )
        self.E1 = NumParam(info='First saturation point',
                           tex_name='E_1',
                           default=0.0,
                           unit='p.u.',
                           )
        self.SE1 = NumParam(info='Value at first saturation point',
                            tex_name='S_{E1}',
                            default=0.0,
                            unit='p.u.',
                            )
        self.E2 = NumParam(info='Second saturation point',
                           tex_name='E_2',
                           default=1.0,
                           unit='p.u.',
                           )
        self.SE2 = NumParam(info='Value at second saturation point',
                            tex_name='S_{E2}',
                            default=1.0,
                            unit='p.u.',
                            )


class IEEET1Model(ExcBase):
    """
    IEEE Type 1 exciter model.
    """

    def __init__(self, system, config):
        ExcBase.__init__(self, system, config)

        # Set VRMAX to 999 when VRMAX = 0
        self._zVRM = FlagValue(self.VRMAX, value=0,
                               tex_name='z_{VRMAX}',
                               )
        self.VRMAXc = ConstService(v_str='VRMAX + 999*(1-_zVRM)',
                                   info='Set VRMAX=999 when zero',
                                   )
        # Saturation
        self.SAT = ExcQuadSat(self.E1, self.SE1, self.E2, self.SE2,
                              info='Field voltage saturation',
                              )

        self.Se0 = ConstService(info='Initial saturation output',
                                tex_name='S_{e0}',
                                v_str='Indicator(vf0>SAT_A) * SAT_B * (SAT_A - vf0) ** 2',
                                )
        self.vr0 = ConstService(info='Initial vr',
                                tex_name='V_{r0}',
                                v_str='KE * vf0 + Se0')
        self.vb0 = ConstService(info='Initial vb',
                                tex_name='V_{b0}',
                                v_str='vr0 / KA')
        self.vfe0 = ConstService(v_str='vf0 * KE + Se0',
                                 tex_name='V_{FE0}',
                                 )

        self.vref = Algeb(info='Reference voltage input',
                          tex_name='V_{ref}',
                          unit='p.u.',
                          v_str='v + vb0',
                          e_str='vref0 - vref'
                          )

        self.vref0 = PostInitService(info='Const reference voltage',
                                     tex_name='V_{ref0}',
                                     v_str='vref',
                                     )

        self.LG = Lag(u=self.v, T=self.TR, K=1,
                      info='Sensing delay',
                      )
        # NOTE: for offline exciters, `vi` equation ignores ext. voltage changes
        self.vi = Algeb(info='Total input voltages',
                        tex_name='V_i',
                        unit='p.u.',
                        e_str='ue * (-LG_y + vref - vi)',
                        v_str='-v + vref',
                        diag_eps=True,
                        )
        self.LA = LagAntiWindup(u='ue * (vi - WF_y)',
                                T=self.TA,
                                K=self.KA,
                                upper=self.VRMAXc,
                                lower=self.VRMIN,
                                info='Anti-windup lag',
                                )
        self.VFE = Algeb(info='Combined saturation feedback',
                         tex_name='V_{FE}',
                         unit='p.u.',
                         v_str='vfe0',
                         e_str='ue * (INT_y * KE + Se - VFE)',
                         diag_eps=True,
                         )

        self.INT = Integrator(u='ue * (LA_y - VFE)',
                              T=self.TE,
                              K=1,
                              y0=self.vf0,
                              info='Integrator',
                              )

        self.SL = LessThan(u=self.vout, bound=self.SAT_A, equal=False, enable=True, cache=False)

        self.Se = Algeb(tex_name=r"S_e(|V_{out}|)", info='saturation output',
                        v_str='Se0',
                        e_str='SL_z0 * (INT_y - SAT_A) ** 2 * SAT_B - Se',
                        )

        self.WF = Washout(u=self.vout, T=self.TF, K=self.KF, info='Stablizing circuit feedback')

        self.vout.e_str = 'ue * (INT_y - vout)'


class IEEET1(IEEET1Data, IEEET1Model):
    def __init__(self, system, config):
        IEEET1Data.__init__(self)
        IEEET1Model.__init__(self, system, config)
